# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""`UNet network architecture <unet_>`_, from :cite:p:`navab_u-net_2015`."""

import logging
import typing

import torch.nn
import torch.utils.data

from ...data.typing import TransformSequence
from ..losses import SoftJaccardAndBCEWithLogitsLoss
from .backbones.vgg import vgg16_for_segmentation
from .make_layers import UnetBlock, conv_with_kaiming_uniform
from .model import Model

logger = logging.getLogger(__name__)


class UNetHead(torch.nn.Module):
    """UNet head module.

    Parameters
    ----------
    in_channels_list
        Number of channels for each feature map that is returned from backbone.
    pixel_shuffle
        If True, upsample using PixelShuffleICNR.
    """

    def __init__(self, in_channels_list: list[int], pixel_shuffle=False):
        super().__init__()
        # number of channels
        c_decode1, c_decode2, c_decode3, c_decode4, c_decode5 = in_channels_list

        # build layers
        self.decode4 = UnetBlock(c_decode5, c_decode4, pixel_shuffle, middle_block=True)
        self.decode3 = UnetBlock(c_decode4, c_decode3, pixel_shuffle)
        self.decode2 = UnetBlock(c_decode3, c_decode2, pixel_shuffle)
        self.decode1 = UnetBlock(c_decode2, c_decode1, pixel_shuffle)
        self.final = conv_with_kaiming_uniform(c_decode1, 1, 1)

    def forward(self, x: list[torch.Tensor]):
        """Forward pass.

        Parameters
        ----------
        x
            List of tensors as returned from the backbone network.
            First element: height and width of input image.
            Remaining elements: feature maps for each feature level.

        Returns
        -------
            OUtput of the forward pass.
        """
        # NOTE: x[0]: height and width of input image not needed in U-Net architecture
        decode4 = self.decode4(x[5], x[4])
        decode3 = self.decode3(decode4, x[3])
        decode2 = self.decode2(decode3, x[2])
        decode1 = self.decode1(decode2, x[1])
        return self.final(decode1)


class Unet(Model):
    """`UNet network architecture <unet_>`_, from :cite:p:`navab_u-net_2015`.

    Parameters
    ----------
    loss_type
        The loss to be used for training and evaluation.

        .. warning::

           The loss should be set to always return batch averages (as opposed
           to the batch sum), as our logging system expects it so.
    loss_arguments
        Arguments to the loss.
    optimizer_type
        The type of optimizer to use for training.
    optimizer_arguments
        Arguments to the optimizer after ``params``.
    scheduler_type
        The type of scheduler to use for training.
    scheduler_arguments
        Arguments to the scheduler after ``params``.
    model_transforms
        An optional sequence of torch modules containing transforms to be
        applied on the input **before** it is fed into the network.
    augmentation_transforms
        An optional sequence of torch modules containing transforms to be
        applied on the input **before** it is fed into the network.
    pretrained
        If True, will use VGG16 pretrained weights.
    """

    def __init__(
        self,
        loss_type: type[torch.nn.Module] = SoftJaccardAndBCEWithLogitsLoss,
        loss_arguments: dict[str, typing.Any] | None = None,
        optimizer_type: type[torch.optim.Optimizer] = torch.optim.Adam,
        optimizer_arguments: dict[str, typing.Any] | None = None,
        scheduler_type: type[torch.optim.lr_scheduler.LRScheduler] | None = None,
        scheduler_arguments: dict[str, typing.Any] | None = None,
        model_transforms: TransformSequence | None = None,
        augmentation_transforms: TransformSequence | None = None,
        pretrained: bool = False,
    ):
        super().__init__(
            name="unet",
            loss_type=loss_type,
            loss_arguments=loss_arguments,
            optimizer_type=optimizer_type,
            optimizer_arguments=optimizer_arguments,
            scheduler_type=scheduler_type,
            scheduler_arguments=scheduler_arguments,
            model_transforms=model_transforms,
            augmentation_transforms=augmentation_transforms,
            num_classes=1,  # fixed at current implementation
        )

        if pretrained:
            from ..normalizer import make_imagenet_normalizer

            Model.normalizer.fset(self, make_imagenet_normalizer())  # type: ignore[attr-defined]

        self.backbone = vgg16_for_segmentation(
            pretrained=pretrained,
            return_features=[3, 8, 14, 22, 29],
        )

        self.head = UNetHead([64, 128, 256, 512, 512], pixel_shuffle=False)

    def forward(self, x):
        x = self.normalizer(x)
        x = self.backbone(x)
        return self.head(x)



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""Elastic deformation with 80% probability.

This sole data augmentation was proposed by Pasa in the article "Efficient Deep
Network Architectures for Fast Chest X-Ray Tuberculosis Screening and
Visualization".

Reference: :cite:p:`pasa_efficient_2019`
"""

from ...data.augmentations import ElasticDeformation

augmentations = [ElasticDeformation(p=0.8)]



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""Custom lightning loggers."""

import os
import typing

from lightning.fabric.utilities.types import _PATH
from lightning.pytorch.loggers import TensorBoardLogger


class CustomTensorboardLogger(TensorBoardLogger):
    r"""Custom implementation implementation of lightning's TensorboardLogger.

    This implementation puts all logs inside the same directory, instead of a
    separate "version_n" directories, which is the default lightning behaviour.

    Parameters
    ----------
    save_dir
        Directory where to save the logs to.
    name
        Experiment name. Defaults to ``default``. If it is the empty string
        then no per-experiment subdirectory is used.
    version
        Experiment version. If version is not specified the logger inspects the
        save directory for existing versions, then automatically assigns the
        next available version. If it is a string then it is used as the
        run-specific subdirectory name, otherwise ``version_${version}`` is
        used.
    log_graph
        Adds the computational graph to tensorboard. This requires that the
        user has defined the `self.example_input_array` attribute in their
        model.
    default_hp_metric
        Enables a placeholder metric with key `hp_metric` when
        `log_hyperparams` is called without a metric (otherwise calls to
        log_hyperparams without a metric are ignored).
    prefix
        A string to put at the beginning of metric keys.
    sub_dir
        Sub-directory to group TensorBoard logs. If a sub_dir argument is
        passed then logs are saved in ``/save_dir/name/version/sub_dir/``.
        Defaults to ``None`` in which logs are saved in
        ``/save_dir/name/version/``.
    \**kwargs
        Additional arguments used by :py:class:`tensorboardX.SummaryWriter` can
        be passed as keyword arguments in this logger. To automatically flush
        to disk, ``max_queue`` sets the size of the queue for pending logs before
        flushing. ``flush_secs`` determines how many seconds elapses before
        flushing.
    """

    def __init__(
        self,
        save_dir: _PATH,
        name: str = "lightning-logs",
        version: int | str | None = None,
        log_graph: bool = False,
        default_hp_metric: bool = True,
        prefix: str = "",
        sub_dir: _PATH | None = None,
        **kwargs: dict[str, typing.Any],
    ):
        super().__init__(
            save_dir,
            name,
            version,
            log_graph,
            default_hp_metric,
            prefix,
            sub_dir,
            **kwargs,
        )

    @property
    def log_dir(self) -> str:
        return os.path.join(self.save_dir, self.name)  # noqa: PTH118



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import pathlib
import typing

import click
from clapper.click import ResourceOption, verbosity_option

from ....models.classify.typing import SaliencyMapAlgorithm
from ...click import ConfigCommand
from ...logging import setup_cli_logger

logger = setup_cli_logger()


@click.command(
    entry_point_group="mednet.config",
    cls=ConfigCommand,
    epilog="""Examples:

1. Generate saliency maps using GradCAM for all prediction dataloaders on a
   DataModule, using a pre-trained DenseNet model, and saves them as
   numpy-pickeled objects on the output directory:

   .. code:: sh

      mednet classify saliency generate -vv densenet tbx11k-v1-healthy-vs-atb --saliency-map-algorithm="gradcam" --weight=path/to/model-at-lowest-validation-loss.ckpt --output-folder=path/to/output

""",
)
@click.option(
    "--model",
    "-m",
    help="""A lightning module instance implementing the network architecture
    (not the weights, necessarily) to be used for inference.  Currently, only
    supports pasa and densenet models.""",
    required=True,
    cls=ResourceOption,
)
@click.option(
    "--datamodule",
    "-d",
    help="""A lightning DataModule that will be asked for prediction data
    loaders. Typically, this includes all configured splits in a DataModule,
    however this is not a requirement.  A DataModule that returns a single
    dataloader for prediction (wrapped in a dictionary) is acceptable.""",
    required=True,
    cls=ResourceOption,
)
@click.option(
    "--output-folder",
    "-o",
    help="Directory in which to store saliency maps (created if does not exist)",
    required=True,
    type=click.Path(
        exists=False,
        file_okay=False,
        dir_okay=True,
        writable=True,
        path_type=pathlib.Path,
    ),
    default="saliency-maps",
    show_default=True,
    cls=ResourceOption,
)
@click.option(
    "--device",
    "-x",
    help='A string indicating the device to use (e.g. "cpu" or "cuda:0")',
    show_default=True,
    required=True,
    default="cpu",
    cls=ResourceOption,
)
@click.option(
    "--cache-samples/--no-cache-samples",
    help="If set to True, loads the sample into memory, "
    "otherwise loads them at runtime.",
    required=True,
    show_default=True,
    default=False,
    cls=ResourceOption,
)
@click.option(
    "--weight",
    "-w",
    help="""Path or URL to a pretrained model file (`.ckpt` extension),
    corresponding to the architecture set with `--model`.  Optionally, you may
    also pass a directory containing the result of a training session, in which
    case either the best (lowest validation) or latest model will be loaded.""",
    required=True,
    cls=ResourceOption,
    type=click.Path(
        exists=True,
        file_okay=True,
        dir_okay=True,
        readable=True,
        path_type=pathlib.Path,
    ),
)
@click.option(
    "--parallel",
    "-P",
    help="""Use multiprocessing for data loading: if set to -1 (default),
    disables multiprocessing data loading.  Set to 0 to enable as many data
    loading instances as processing cores available in the system.  Set to
    >= 1 to enable that many multiprocessing instances for data loading.""",
    type=click.IntRange(min=-1),
    show_default=True,
    required=True,
    default=-1,
    cls=ResourceOption,
)
@click.option(
    "--saliency-map-algorithm",
    "-s",
    help="""Saliency map algorithm to be used.""",
    type=click.Choice(
        typing.get_args(SaliencyMapAlgorithm),
        case_sensitive=False,
    ),
    default="gradcam",
    show_default=True,
    cls=ResourceOption,
)
@click.option(
    "--target-class",
    "-C",
    help="""This option should only be used with multiclass models.  It
    defines the class to target for saliency estimation. Can be either set to
    "all" or "highest". "highest" (the default), means only saliency maps for
    the class with the highest activation will be generated.""",
    required=False,
    type=click.Choice(
        ["highest", "all"],
        case_sensitive=False,
    ),
    default="highest",
    cls=ResourceOption,
)
@click.option(
    "--positive-only/--no-positive-only",
    "-z/-Z",
    help="""If set, and the model chosen has a single output (binary), then
    saliency maps will only be generated for samples of the positive class.
    This option has no effect for multiclass models.""",
    default=False,
    cls=ResourceOption,
)
@click.option(
    "--only-dataset",
    "-S",
    help="""If set, will only run the command for the named dataset on the
    provided datamodule, skipping any other dataset.""",
    cls=ResourceOption,
)
@verbosity_option(logger=logger, expose_value=False)
def generate(
    model,
    datamodule,
    output_folder,
    device,
    cache_samples,
    weight,
    parallel,
    saliency_map_algorithm,
    target_class,
    positive_only,
    only_dataset,
    **_,
) -> None:  # numpydoc ignore=PR01
    """Generate saliency maps for locations on input images that affected the
    prediction.

    The quality of saliency information depends on the saliency map
    algorithm and trained model.
    """

    from ....engine.classify.saliency.generator import run
    from ....engine.device import DeviceManager
    from ....engine.trainer import validate_model_datamodule
    from ....models.classify.model import Model as ClassificationModel
    from ....utils.checkpointer import get_checkpoint_to_run_inference
    from ...utils import save_json_metadata

    validate_model_datamodule(model, datamodule)
    assert isinstance(model, ClassificationModel)

    logger.info(f"Output folder: {output_folder}")
    output_folder.mkdir(parents=True, exist_ok=True)

    device_manager = DeviceManager(device)

    # sets-up the data module
    datamodule.cache_samples = cache_samples
    datamodule.model_transforms = list(model.model_transforms)
    datamodule.batch_size = 1
    datamodule.parallel = -1
    datamodule.prepare_data()
    datamodule.setup(stage="predict")

    if weight.is_dir():
        weight = get_checkpoint_to_run_inference(weight)

    logger.info(f"Loading checkpoint from `{weight}`...")
    model = type(model).load_from_checkpoint(weight, strict=False)

    save_json_metadata(
        output_file=output_folder / "generation.meta.json",
        datamodule=datamodule,
        model=model,
        output_folder=output_folder,
        device=device,
        cache_samples=cache_samples,
        weight=weight,
        parallel=parallel,
        saliency_map_algorithm=saliency_map_algorithm,
        target_class=target_class,
        positive_only=positive_only,
        only_dataset=only_dataset,
    )

    run(
        model=model,
        datamodule=datamodule,
        device_manager=device_manager,
        saliency_map_algorithm=saliency_map_algorithm,
        target_class=target_class,
        positive_only=positive_only,
        output_folder=output_folder,
        only_dataset=only_dataset,
    )



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import pathlib
import typing

import click
from clapper.click import ResourceOption, verbosity_option

from ...engine.segment.evaluator import SUPPORTED_METRIC_TYPE
from ..click import ConfigCommand
from ..logging import setup_cli_logger

# avoids X11/graphical desktop requirement when creating plots
__import__("matplotlib").use("agg")

logger = setup_cli_logger()


@click.command(
    entry_point_group="mednet.config",
    cls=ConfigCommand,
    epilog="""Examples:

\b
  1. Runs evaluation from existing predictions:

     .. code:: sh

        $ mednet segment evaluate -vv --predictions=path/to/predictions.json --output-folder=path/to/results

  2. Runs evaluation from existing predictions, incorporate a second set of
     annotations into the evaluation:

     .. code:: sh

        $ mednet segment evaluate -vv --predictions=path/to/predictions.json --compare-annotator=path/to/annotations.json --output-folder=path/to/results
""",
)
@click.option(
    "--predictions",
    "-p",
    help="""Path to the JSON file describing available predictions. The actual
    predictions are supposed to lie on the same folder.""",
    required=True,
    type=click.Path(
        file_okay=True,
        dir_okay=False,
        writable=False,
        path_type=pathlib.Path,
    ),
    cls=ResourceOption,
)
@click.option(
    "--output-folder",
    "-o",
    help="Directory in which to store results (created if does not exist)",
    required=True,
    type=click.Path(
        file_okay=False,
        dir_okay=True,
        writable=True,
        path_type=pathlib.Path,
    ),
    default="results",
    show_default=True,
    cls=ResourceOption,
)
@click.option(
    "--threshold",
    "-t",
    help="""This number is used to define positives and negatives from
    probability maps, and used to report metrics based on a threshold chosen *a
    priori*. It can be set to a floating-point value, or to the name of dataset
    split in ``--predictions``.
    """,
    default="0.5",
    show_default=True,
    required=False,
    cls=ResourceOption,
)
@click.option(
    "--metric",
    "-m",
    help="""If threshold is set to the name of a split in ``--predictions``,
    then this parameter defines the metric function to be used to evaluate the
    threshold at which the metric reaches its maximum value. All other splits
    are evaluated with respect to this threshold.""",
    default="f1",
    type=click.Choice(typing.get_args(SUPPORTED_METRIC_TYPE), case_sensitive=True),
    show_default=True,
    required=True,
    cls=ResourceOption,
)
@click.option(
    "--steps",
    "-s",
    help="""Number of steps for evaluating metrics on various splits. This
    value is used when drawing precision-recall plots, or when deciding the
    highest metric value on splits.""",
    default=100,
    type=click.IntRange(10),
    show_default=True,
    required=True,
    cls=ResourceOption,
)
@click.option(
    "--compare-annotator",
    "-a",
    help="""Path to a JSON file as produced by the CLI ``dump-annotations``,
    containing splits and sample lists with associated HDF5 files where we can
    find pre-processed annotation masks.  These annotations will be compared
    with the target annotations on the main predictions.  In this case, a row
    is added for each available split in the evaluation table.""",
    required=False,
    default=None,
    type=click.Path(
        file_okay=True,
        dir_okay=False,
        writable=False,
        path_type=pathlib.Path,
    ),
    cls=ResourceOption,
)
@click.option(
    "--plot/--no-plot",
    "-P",
    help="""If set, then also produces figures containing the plots of
    performance curves and score histograms.""",
    required=True,
    show_default=True,
    default=True,
    cls=ResourceOption,
)
@verbosity_option(logger=logger, expose_value=False)
def evaluate(
    predictions: pathlib.Path,
    output_folder: pathlib.Path,
    threshold: str | float,
    metric: SUPPORTED_METRIC_TYPE,
    steps: int,
    compare_annotator: pathlib.Path,
    plot: bool,
    **_,  # ignored
):  # numpydoc ignore=PR01
    """Evaluate predictions (from a model) on a segmentation task."""

    import matplotlib.backends.backend_pdf

    from ...engine.segment.evaluator import (
        compare_annotators,
        make_plots,
        make_table,
        run,
    )
    from ..utils import save_json_metadata, save_json_with_backup

    evaluation_file = output_folder / "evaluation.json"

    save_json_metadata(
        output_file=evaluation_file.with_suffix(".meta.json"),
        predictions=str(predictions),
        output_folder=str(output_folder),
        threshold=threshold,
        metric=metric,
        steps=steps,
        compare_annotator=str(compare_annotator),
        plot=plot,
    )

    eval_json_data, threshold_value = run(predictions, steps, threshold, metric)

    if compare_annotator is not None:
        logger.info(f"Comparing 2nd. annotator using `{str(compare_annotator)}`...")
        comparison = compare_annotators(predictions, compare_annotator)
        assert comparison, (
            f"No matching split-names found on `{str(compare_annotator)}`. "
            f"Are these the right annotations?"
        )
        for split_name, values in comparison.items():
            eval_json_data.setdefault(split_name, {})["second_annotator"] = values

    # Records full result analysis to a JSON file
    logger.info(f"Saving evaluation results at `{str(evaluation_file)}`...")
    save_json_with_backup(evaluation_file, eval_json_data)

    # Produces and records table
    table = make_table(eval_json_data, threshold_value, "rst")
    click.echo(table)

    output_table = evaluation_file.with_suffix(".rst")
    logger.info(f"Saving tabulated performance summary at `{str(output_table)}`...")
    output_table.parent.mkdir(parents=True, exist_ok=True)
    with output_table.open("w") as f:
        f.write(table)

    # Plots pre-calculated curves, if the user asked to do so.
    if plot:
        figure_path = evaluation_file.with_suffix(".pdf")
        logger.info(f"Saving evaluation figures at `{str(figure_path)}`...")
        with matplotlib.backends.backend_pdf.PdfPages(figure_path) as pdf:
            for fig in make_plots(eval_json_data):
                pdf.savefig(fig)



# SPDX-FileCopyrightText: Copyright © 2024 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""IOSTAR (training set) for vessel and optic-disc segmentation.

The IOSTAR vessel segmentation dataset includes 30 images with a resolution of
1024×1024 pixels. All the vessels in this dataset are annotated by a group of
experts working in the field of retinal image analysis. Additionally the
dataset includes annotations for the optic disc and the artery/vein ratio.

* Reference: :cite:p:`zhang_robust_2016`
* Original resolution (height x width): 1024×1024
* Split reference: :cite:p:`meyer_deep_2017`
* Protocol ``vessel``:

  * Training samples: 20 (including labels and masks)
  * Test samples: 10 (including labels and masks)

* Protocol ``optic-disc``:

  * Training samples: 20 (including labels and masks)
  * Test samples: 10 (including labels and masks)

This module contains the base declaration of common data modules and raw-data
loaders for this database. All configured splits inherit from this definition.
"""

import importlib.resources.abc
import os
import pathlib
import typing

import PIL.Image
import torch
from torchvision import tv_tensors
from torchvision.transforms.v2.functional import to_dtype, to_image

from ...models.transforms import crop_image_to_mask
from ...utils.rc import load_rc
from ..datamodule import CachingDataModule
from ..split import JSONDatabaseSplit
from ..typing import RawDataLoader as BaseDataLoader
from ..typing import Sample

DATABASE_SLUG = __name__.rsplit(".", 1)[-1]
"""Pythonic name to refer to this database."""

CONFIGURATION_KEY_DATADIR = "datadir." + DATABASE_SLUG
"""Key to search for in the configuration file for the root directory of this
database."""


class RawDataLoader(BaseDataLoader):
    """A specialized raw-data-loader for the iostar dataset."""

    datadir: pathlib.Path
    """This variable contains the base directory where the database raw data is
    stored."""

    def __init__(self):
        self.datadir = pathlib.Path(
            load_rc().get(CONFIGURATION_KEY_DATADIR, os.path.realpath(os.curdir))
        )

    def sample(self, sample: typing.Any) -> Sample:
        """Load a single image sample from the disk.

        Parameters
        ----------
        sample
            A tuple containing the path suffix, within the dataset root folder,
            where to find the image to be loaded, and an integer, representing the
            sample label.

        Returns
        -------
            The sample representation.
        """

        image = PIL.Image.open(self.datadir / sample[0]).convert(mode="RGB")
        image = to_dtype(to_image(image), torch.float32, scale=True)

        target = self.target(sample)

        mask = PIL.Image.open(self.datadir / sample[2]).convert(mode="1", dither=None)
        mask = to_dtype(to_image(mask), torch.float32, scale=True)

        image = tv_tensors.Image(crop_image_to_mask(image, mask))
        target = tv_tensors.Mask(crop_image_to_mask(target, mask))
        mask = tv_tensors.Mask(mask)

        return dict(image=image, target=target, mask=mask, name=sample[0])

    def target(self, sample: typing.Any) -> torch.Tensor:
        """Load only sample target from its raw representation.

        Parameters
        ----------
        sample
            A tuple containing the path suffix, within the dataset root folder,
            where to find the image to be loaded, and an integer, representing
            the sample target.

        Returns
        -------
            The label corresponding to the specified sample, encapsulated as a
            torch float tensor.
        """

        target = PIL.Image.open(self.datadir / sample[1]).convert(mode="1", dither=None)
        return to_dtype(to_image(target), torch.float32, scale=True)


class DataModule(CachingDataModule):
    """IOSTAR (training set) for vessel and optic-disc segmentation.

    Parameters
    ----------
    split_path
        Path or traversable (resource) with the JSON split description to load.
    """

    def __init__(self, split_path: pathlib.Path | importlib.resources.abc.Traversable):
        super().__init__(
            database_split=JSONDatabaseSplit(split_path),
            raw_data_loader=RawDataLoader(),
            database_name=DATABASE_SLUG,
            split_name=split_path.name.rsplit(".", 2)[0],
            task="segmentation",
            num_classes=1,
        )



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""Tests for our CLI applications."""

import contextlib
import re

import pytest
from click.testing import CliRunner


@contextlib.contextmanager
def stdout_logging():
    # copy logging messages to std out

    import io
    import logging

    buf = io.StringIO()
    ch = logging.StreamHandler(buf)
    ch.setFormatter(logging.Formatter("%(message)s"))
    ch.setLevel(logging.INFO)
    logger = logging.getLogger("mednet")
    logger.addHandler(ch)
    yield buf
    logger.removeHandler(ch)


def _assert_exit_0(result):
    assert (
        result.exit_code == 0
    ), f"Exit code {result.exit_code} != 0 -- Output:\n{result.output}"


def _check_help(entry_point):
    runner = CliRunner()
    result = runner.invoke(entry_point, ["--help"])
    _assert_exit_0(result)
    assert result.output.startswith("Usage:")


@pytest.mark.skip_if_rc_var_not_set("datadir.drive")
def test_config_describe_drive():
    from mednet.scripts.config import describe

    runner = CliRunner()
    result = runner.invoke(describe, ["drive"])
    _assert_exit_0(result)
    assert (
        ":py:mod:`DRIVE database <mednet.data.segment.drive>` (default split)."
        in result.output
    )


@pytest.mark.skip_if_rc_var_not_set("datadir.drive")
def test_database_check():
    from mednet.scripts.database import check

    runner = CliRunner()
    result = runner.invoke(check, ["--verbose", "--limit=1", "drive"])
    _assert_exit_0(result)


def test_main_help():
    from mednet.scripts.segment.cli import segment

    _check_help(segment)


def _str_counter(substr, s):
    return sum(1 for _ in re.finditer(substr, s, re.MULTILINE))


def test_evaluate_help():
    from mednet.scripts.segment.evaluate import evaluate

    _check_help(evaluate)


def test_dump_annotations_help():
    from mednet.scripts.segment.dump_annotations import dump_annotations

    _check_help(dump_annotations)


def test_view_help():
    from mednet.scripts.segment.view import view

    _check_help(view)


@pytest.mark.slow
@pytest.mark.skip_if_rc_var_not_set("datadir.drive")
def test_train_lwnet_drive(session_tmp_path):
    from mednet.scripts.train import train
    from mednet.utils.checkpointer import (
        CHECKPOINT_EXTENSION,
        _get_checkpoint_from_alias,
    )

    runner = CliRunner()

    with stdout_logging() as buf:
        output_folder = session_tmp_path / "segmentation-standalone"
        result = runner.invoke(
            train,
            [
                "lwnet",
                "drive",
                "-vv",
                "--epochs=1",
                "--batch-size=1",
                f"--output-folder={str(output_folder)}",
            ],
        )
        _assert_exit_0(result)

        # asserts checkpoints are there, or raises FileNotFoundError
        last = _get_checkpoint_from_alias(output_folder, "periodic")
        assert last.name.endswith("epoch=0" + CHECKPOINT_EXTENSION)
        best = _get_checkpoint_from_alias(output_folder, "best")
        assert best.name.endswith("epoch=0" + CHECKPOINT_EXTENSION)

        assert len(list((output_folder / "logs").glob("events.out.tfevents.*"))) == 1
        assert (output_folder / "train.meta.json").exists()

        keywords = {
            r"^Writing run metadata at .*$": 1,
            r"^Loading dataset:`train` without caching. Trade-off: CPU RAM usage: less | Disk I/O: more.$": 1,
            # Starting from scratch:
            # 1) it is a known model, so the next message should NOT appear
            r"^Model `.*` is not known. Skipping input normalization from train dataloader setup (unsupported external model).$": 0,
            # 2) should compute a new set of normalization factors
            r"^Computing z-norm input normalization from dataloader.$": 1,
            # 3) should NOT reload pre-calculated normalization factors from checkpoint
            r"^Restored input normalizer from checkpoint.$": 0,
            # The loss used in LWNet balances batches on-the-fly so pre-balancing is not
            # supported
            r"^Loss `.*` is not supported and will not be balanced.$": 1,
            r"^Applying train/valid loss balancing...$": 0,
            r"^Training for at most 1 epochs.$": 1,
            r"^Dataset `train` is already setup. Not re-instantiating it.$": 3,
        }
        buf.seek(0)
        logging_output = buf.read()

        for k, v in keywords.items():
            assert _str_counter(k, logging_output) == v, (
                f"Count for string '{k}' appeared "
                f"({_str_counter(k, logging_output)}) "
                f"instead of the expected {v}:\nOutput:\n{logging_output}"
            )


@pytest.mark.slow
@pytest.mark.skip_if_rc_var_not_set("datadir.drive")
def test_predict_lwnet_drive(session_tmp_path):
    from mednet.scripts.predict import predict
    from mednet.utils.checkpointer import (
        CHECKPOINT_EXTENSION,
        _get_checkpoint_from_alias,
    )

    runner = CliRunner()

    with stdout_logging() as buf:
        output_folder = session_tmp_path / "segmentation-standalone"
        last_ckpt = _get_checkpoint_from_alias(output_folder, "periodic")
        assert last_ckpt.name.endswith("epoch=0" + CHECKPOINT_EXTENSION)
        result = runner.invoke(
            predict,
            [
                "lwnet",
                "drive",
                "-vv",
                "--batch-size=1",
                f"--weight={str(last_ckpt)}",
                f"--output-folder={str(output_folder)}",
            ],
        )
        _assert_exit_0(result)

        assert (output_folder / "predictions.meta.json").exists()
        assert (output_folder / "predictions.json").exists()

        keywords = {
            r"^Writing run metadata at .*$": 1,
            r"^Loading dataset: * without caching. Trade-off: CPU RAM usage: less | Disk I/O: more$": 2,
            r"^Loading checkpoint from .*$": 1,
            # Prediction
            # 1) should NOT compute a new set of normalization factors
            r"^Computing z-norm input normalization from dataloader.$": 0,
            # 2) should reload pre-calculated normalization factors from checkpoint
            r"^Restored input normalizer from checkpoint.$": 1,
            r"^Running prediction on `train` split...$": 1,
            r"^Running prediction on `test` split...$": 1,
            r"^Predictions saved to .*$": 1,
        }

        buf.seek(0)
        logging_output = buf.read()

        for k, v in keywords.items():
            assert _str_counter(k, logging_output) == v, (
                f"Count for string '{k}' appeared "
                f"({_str_counter(k, logging_output)}) "
                f"instead of the expected {v}:\nOutput:\n{logging_output}"
            )


@pytest.mark.slow
@pytest.mark.skip_if_rc_var_not_set("datadir.drive")
def test_dump_annotations_lwnet_drive(session_tmp_path):
    from mednet.scripts.segment.dump_annotations import dump_annotations

    runner = CliRunner()

    with stdout_logging() as buf:
        output_folder = (
            session_tmp_path / "segmentation-standalone" / "second-annotator"
        )
        result = runner.invoke(
            dump_annotations,
            [
                "lwnet",
                "drive-2nd",
                "-vv",
                f"--output-folder={str(output_folder)}",
            ],
        )
        _assert_exit_0(result)

        assert (output_folder / "annotations.meta.json").exists()
        assert (output_folder / "annotations.json").exists()

        keywords = {
            r"^Writing run metadata at .*$": 1,
            r"^Loading dataset:.*$": 1,
            r"^Dumping annotations from split.*$": 1,
        }

        buf.seek(0)
        logging_output = buf.read()

        for k, v in keywords.items():
            assert _str_counter(k, logging_output) == v, (
                f"Count for string '{k}' appeared "
                f"({_str_counter(k, logging_output)}) "
                f"instead of the expected {v}:\nOutput:\n{logging_output}"
            )


@pytest.mark.slow
@pytest.mark.skip_if_rc_var_not_set("datadir.drive")
def test_evaluate_lwnet_drive(session_tmp_path):
    from mednet.scripts.segment.evaluate import evaluate

    runner = CliRunner()

    with stdout_logging() as buf:
        output_folder = session_tmp_path / "segmentation-standalone"
        result = runner.invoke(
            evaluate,
            [
                "-vv",
                f"--predictions={str(output_folder / 'predictions.json')}",
                f"--output-folder={str(output_folder)}",
                "--threshold=test",
                f"--compare-annotator={str(output_folder / 'second-annotator' / 'annotations.json')}",
            ],
        )
        _assert_exit_0(result)

        assert (output_folder / "evaluation.json").exists()
        assert (output_folder / "evaluation.meta.json").exists()
        assert (output_folder / "evaluation.rst").exists()
        assert (output_folder / "evaluation.pdf").exists()

        keywords = {
            r"^Writing run metadata at .*$": 1,
            r"^Counting true/false positive/negatives at split.*$": 2,
            r"^Evaluating threshold on split .*$": 1,
            r"^Computing metrics on split .*$": 2,
            r"^Comparing 2nd. annotator using .*$": 1,
            r"^Saving evaluation results at .*$": 1,
            r"^Saving tabulated performance summary at .*$": 1,
            r"^Saving evaluation figures at .*$": 1,
        }
        buf.seek(0)
        logging_output = buf.read()

        for k, v in keywords.items():
            assert _str_counter(k, logging_output) == v, (
                f"Count for string '{k}' appeared "
                f"({_str_counter(k, logging_output)}) "
                f"instead of the expected {v}:\nOutput:\n{logging_output}"
            )


@pytest.mark.slow
@pytest.mark.skip_if_rc_var_not_set("datadir.drive")
def test_view_lwnet_drive(session_tmp_path):
    from mednet.scripts.segment.view import view

    runner = CliRunner()

    with stdout_logging() as buf:
        output_folder = session_tmp_path / "segmentation-standalone"
        result = runner.invoke(
            view,
            [
                "-vv",
                f"--predictions={str(output_folder / 'predictions.json')}",
                f"--output-folder={str(output_folder)}",
            ],
        )
        _assert_exit_0(result)

        assert len(list((output_folder / "training" / "images").glob("*.png"))) == 20
        assert len(list((output_folder / "test" / "images").glob("*.png"))) == 20

        keywords = {
            r"^Creating 20 visualisations for split.*$": 2,
            r"^Set --threshold.*$": 1,
        }
        buf.seek(0)
        logging_output = buf.read()

        for k, v in keywords.items():
            assert _str_counter(k, logging_output) == v, (
                f"Count for string '{k}' appeared "
                f"({_str_counter(k, logging_output)}) "
                f"instead of the expected {v}:\nOutput:\n{logging_output}"
            )


@pytest.mark.slow
@pytest.mark.skip_if_rc_var_not_set("datadir.drive")
def test_train_lwnet_drive_from_checkpoint(tmp_path):
    from mednet.scripts.train import train
    from mednet.utils.checkpointer import (
        CHECKPOINT_EXTENSION,
        _get_checkpoint_from_alias,
    )

    runner = CliRunner()

    result0 = runner.invoke(
        train,
        [
            "lwnet",
            "drive",
            "-vv",
            "--epochs=1",
            "--batch-size=1",
            f"--output-folder={str(tmp_path)}",
        ],
    )
    _assert_exit_0(result0)

    # asserts checkpoints are there, or raises FileNotFoundError
    last = _get_checkpoint_from_alias(tmp_path, "periodic")
    assert last.name.endswith("epoch=0" + CHECKPOINT_EXTENSION)
    best = _get_checkpoint_from_alias(tmp_path, "best")
    assert best.name.endswith("epoch=0" + CHECKPOINT_EXTENSION)

    assert (tmp_path / "train.meta.json").exists()
    assert len(list((tmp_path / "logs").glob("events.out.tfevents.*"))) == 1

    with stdout_logging() as buf:
        result = runner.invoke(
            train,
            [
                "lwnet",
                "drive",
                "-vv",
                "--epochs=2",
                "--batch-size=1",
                f"--output-folder={tmp_path}",
            ],
        )
        _assert_exit_0(result)

        # asserts checkpoints are there, or raises FileNotFoundError
        last = _get_checkpoint_from_alias(tmp_path, "periodic")
        assert last.name.endswith("epoch=1" + CHECKPOINT_EXTENSION)
        best = _get_checkpoint_from_alias(tmp_path, "best")

        assert (tmp_path / "train.meta.json").exists()
        assert len(list((tmp_path / "logs").glob("events.out.tfevents.*"))) == 2

        keywords = {
            r"^Writing run metadata at .*$": 1,
            r"^Loading dataset:`train` without caching. Trade-off: CPU RAM usage: less | Disk I/O: more.$": 1,
            # Re-starting from checkpoint:
            # 1) it is a known model, so the next message should NOT appear
            r"^Model `.*` is not known. Skipping input normalization from train dataloader setup (unsupported external model).$": 0,
            # 2) should NOT compute a new set of normalization factors
            r"^Computing z-norm input normalization from dataloader.$": 0,
            # 3) should reload pre-calculated normalization factors from checkpoint
            r"^Restored input normalizer from checkpoint.$": 1,
            # The loss used in LWNet balances batches on-the-fly so pre-balancing is not
            # supported
            r"^Loss `.*` is not supported and will not be balanced.$": 1,
            r"^Applying train/valid loss balancing...$": 0,
            r"^Training for at most 2 epochs.$": 1,
            r"^Resuming from epoch 0 \(checkpoint file: .*$": 1,
            r"^Dataset `train` is already setup. Not re-instantiating it.$": 3,
        }
        buf.seek(0)
        logging_output = buf.read()

        for k, v in keywords.items():
            assert _str_counter(k, logging_output) == v, (
                f"Count for string '{k}' appeared "
                f"({_str_counter(k, logging_output)}) "
                f"instead of the expected {v}:\nOutput:\n{logging_output}"
            )


@pytest.mark.slow
@pytest.mark.skip_if_rc_var_not_set("datadir.drive")
def test_experiment(tmp_path):
    from mednet.scripts.experiment import experiment

    runner = CliRunner()

    num_epochs = 2
    result = runner.invoke(
        experiment,
        [
            "-vv",
            "lwnet",
            "drive",
            f"--epochs={num_epochs}",
            f"--output-folder={str(tmp_path)}",
        ],
    )
    _assert_exit_0(result)

    assert (tmp_path / "train.meta.json").exists()
    assert (tmp_path / f"model-at-epoch={num_epochs-1}.ckpt").exists()

    # Need to glob because we cannot be sure of the checkpoint with lowest validation loss
    assert len(list(tmp_path.glob("model-at-lowest-validation-loss-epoch=*.ckpt"))) == 1
    assert (tmp_path / "trainlog.pdf").exists()
    assert len(list((tmp_path / "logs").glob("events.out.tfevents.*"))) == 1
    assert (tmp_path / "predictions.json").exists()
    assert (tmp_path / "predictions.meta.json").exists()
    assert (tmp_path / "evaluation.json").exists()
    assert (tmp_path / "evaluation.meta.json").exists()
    assert (tmp_path / "evaluation.pdf").exists()
    assert (tmp_path / "evaluation.rst").exists()


@pytest.mark.skip_if_rc_var_not_set("datadir.drive")
def test_preprocess_drive(tmp_path):
    from mednet.scripts.preprocess import preprocess

    runner = CliRunner()

    with stdout_logging() as buf:
        result = runner.invoke(
            preprocess,
            [
                "-vv",
                "lwnet",
                "drive",
                "--limit=3",
                f"--output-folder={str(tmp_path)}",
            ],
        )
        _assert_exit_0(result)

        keywords = {
            r"^Loading dataset:.*$": 2,
            r"^training/images/.*\.tif: \[3, 512, 512\]@torch\.float32$": 3,
            r"^test/images/.*\.tif: \[3, 512, 512\]@torch\.float32$": 3,
        }
        buf.seek(0)
        logging_output = buf.read()

        for k, v in keywords.items():
            assert _str_counter(k, logging_output) == v, (
                f"Count for string '{k}' appeared "
                f"({_str_counter(k, logging_output)}) "
                f"instead of the expected {v}:\nOutput:\n{logging_output}"
            )

        # there are only 58 samples with target = 1
        assert len(list(tmp_path.rglob("**/*.tif"))) == 6  # images
        assert len(list(tmp_path.rglob("**/*.gif"))) == 12  # targets and masks



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""Concrete database-split loaders."""

import functools
import importlib.resources.abc
import json
import logging
import pathlib
import typing

import torch

from ..data.typing import DatabaseSplit, RawDataLoader

logger = logging.getLogger(__name__)


class JSONDatabaseSplit(DatabaseSplit):
    """Define a loader that understands a database split (train, test, etc) in
    JSON format.

    To create a new database split, you need to provide a JSON formatted
    dictionary in a file, with contents similar to the following:

    .. code-block:: json

       {
           "dataset1": [
               [
                   "sample1-data1",
                   "sample1-data2",
                   "sample1-data3",
               ],
               [
                   "sample2-data1",
                   "sample2-data2",
                   "sample2-data3",
               ]
           ],
           "dataset2": [
               [
                   "sample42-data1",
                   "sample42-data2",
                   "sample42-data3",
               ],
           ]
       }

    Your database split many contain any number of (raw) datasets (dictionary
    keys). For simplicity, we recommend to format all sample entries
    similarly so that raw-data-loading is simplified.  Use the function
    :py:func:`check_database_split_loading` to test raw data loading and fine
    tune the dataset split, or its loading.

    Objects of this class behave like a dictionary in which keys are dataset
    names in the split, and values represent samples data and meta-data. The
    actual JSON file descriptors are loaded on demand using
    a py:func:`functools.cached_property`.

    Parameters
    ----------
    path
        Absolute path to a JSON formatted file containing the database split to be
        recognized by this object.
    """

    def __init__(self, path: pathlib.Path | importlib.resources.abc.Traversable):
        self._path = path

    @functools.cached_property
    def _datasets(self) -> DatabaseSplit:
        """Return the DatabaseSplits.

        The first call to this (cached) property will trigger full .json file
        loading from disk.  Subsequent calls will be cached.

        Returns
        -------
        DatabaseSplit
            A dictionary mapping dataset names to lists of JSON objects.
        """

        if str(self._path).endswith(".bz2"):
            logger.debug(f"Loading database split from {str(self._path)}...")
            with __import__("bz2").open(self._path) as f:
                return json.load(f)
        else:
            with self._path.open() as f:
                return json.load(f)

    def __getitem__(self, key: str) -> typing.Sequence[typing.Any]:
        """Access dataset ``key`` from this split."""
        return self._datasets[key]

    def __iter__(self):
        """Iterate over the datasets."""
        return iter(self._datasets)

    def __len__(self) -> int:
        return len(self._datasets)


def check_database_split_loading(
    database_split: DatabaseSplit,
    loader: RawDataLoader,
    limit: int = 0,
) -> int:
    """For each dataset in the split, check if all data can be correctly loaded
    using the provided loader function.

    This function will return the number of errors when loading samples, and will
    log more detailed information to the logging stream.

    Parameters
    ----------
    database_split
        A mapping that contains the database split.  Each key represents the
        name of a dataset in the split.  Each value is a (potentially complex)
        object that represents a single sample.
    loader
        A loader object that knows how to handle full-samples.
    limit
        Maximum number of samples to check (in each split/dataset
        combination) in this dataset.  If set to zero, then check
        everything.

    Returns
    -------
    int
        Number of errors found.
    """

    logger.info(
        "Checking if all samples in all datasets of this split can be loaded...",
    )
    errors = 0
    for dataset, samples in database_split.items():
        samples = samples if not limit else samples[:limit]
        for pos, sample in enumerate(samples):
            try:
                data, _ = loader.sample(sample)
                assert isinstance(data, torch.Tensor)
            except Exception as e:
                logger.info(
                    f"Found error loading entry {pos} in dataset `{dataset}`: {e}",
                )
                errors += 1
    return errors


def make_split(module_name: str, basename: str) -> DatabaseSplit:
    """Return a database split at the provided module.

    This function searches for the database split named ``basename`` at the
    directory where module ``module_name`` is installed, and returns its
    instantiated version.

    Parameters
    ----------
    module_name
        Name of the module where to search for the JSON file.  It should be
        something like ``foo.bar`` for a module defined as
        ``foo/bar/__init__.py``, or ``foo/bar.py``.
    basename
        Name of the .json file containing the split to load.

    Returns
    -------
        An instance of a DatabaseSplit.
    """
    return JSONDatabaseSplit(importlib.resources.files(module_name) / basename)



# SPDX-FileCopyrightText: Copyright © 2024 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
""":py:mod:`REFUGE database <mednet.data.segment.refuge>` (default split for optic-disc segmentation)."""

import importlib.resources

from mednet.data.segment.refuge import DataModule

datamodule = DataModule(
    importlib.resources.files(__package__ or __name__.rsplit(".", 1)[0])
    / "default.json",
    target_type="disc",
)



# SPDX-FileCopyrightText: Copyright © 2024 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
""":py:mod:`CHASE-DB1 database <mednet.data.segment.chasedb1>` (default split with first-annotator labels)."""

import importlib.resources

from mednet.data.segment.chasedb1 import DataModule

datamodule = DataModule(
    importlib.resources.files(__package__ or __name__.rsplit(".", 1)[0])
    / "first-annotator.json"
)



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""Simple 3D convolutional neural network architecture for classification."""

import logging
import typing

import torch
import torch.nn as nn
import torch.nn.functional as F  # noqa: N812
import torch.optim.optimizer
import torch.utils.data

from ...data.typing import TransformSequence
from ..typing import Checkpoint
from .model import Model

logger = logging.getLogger(__name__)


class Conv3DNet(Model):
    """Simple 3D convolutional neural network architecture for classification.

    This network has a linear output.  You should use losses with ``WithLogit``
    instead of cross-entropy versions when training.

    Parameters
    ----------
    loss_type
        The loss to be used for training and evaluation.

        .. warning::

           The loss should be set to always return batch averages (as opposed
           to the batch sum), as our logging system expects it so.
    loss_arguments
        Arguments to the loss.
    optimizer_type
        The type of optimizer to use for training.
    optimizer_arguments
        Arguments to the optimizer after ``params``.
    scheduler_type
        The type of scheduler to use for training.
    scheduler_arguments
        Arguments to the scheduler after ``params``.
    model_transforms
        An optional sequence of torch modules containing transforms to be
        applied on the input **before** it is fed into the network.
    augmentation_transforms
        An optional sequence of torch modules containing transforms to be
        applied on the input **before** it is fed into the network.
    num_classes
        Number of outputs (classes) for this model.
    """

    def __init__(
        self,
        loss_type: type[torch.nn.Module] = torch.nn.BCEWithLogitsLoss,
        loss_arguments: dict[str, typing.Any] = {},
        optimizer_type: type[torch.optim.Optimizer] = torch.optim.Adam,
        optimizer_arguments: dict[str, typing.Any] = {},
        scheduler_type: type[torch.optim.lr_scheduler.LRScheduler] | None = None,
        scheduler_arguments: dict[str, typing.Any] = {},
        model_transforms: TransformSequence = [],
        augmentation_transforms: TransformSequence = [],
        num_classes: int = 1,
    ):
        super().__init__(
            name="cnn3d",
            loss_type=loss_type,
            loss_arguments=loss_arguments,
            optimizer_type=optimizer_type,
            optimizer_arguments=optimizer_arguments,
            scheduler_type=scheduler_type,
            scheduler_arguments=scheduler_arguments,
            model_transforms=model_transforms,
            augmentation_transforms=augmentation_transforms,
            num_classes=num_classes,
        )

        # First convolution block
        self.conv3d_1_1 = nn.Conv3d(
            in_channels=1, out_channels=4, kernel_size=3, stride=1, padding=1
        )
        self.conv3d_1_2 = nn.Conv3d(
            in_channels=4, out_channels=16, kernel_size=3, stride=1, padding=1
        )
        self.conv3d_1_3 = nn.Conv3d(
            in_channels=1, out_channels=16, kernel_size=1, stride=1
        )
        self.batch_norm_1_1 = nn.BatchNorm3d(4)
        self.batch_norm_1_2 = nn.BatchNorm3d(16)
        self.batch_norm_1_3 = nn.BatchNorm3d(16)

        # Second convolution block
        self.conv3d_2_1 = nn.Conv3d(
            in_channels=16, out_channels=24, kernel_size=3, stride=1, padding=1
        )
        self.conv3d_2_2 = nn.Conv3d(
            in_channels=24, out_channels=32, kernel_size=3, stride=1, padding=1
        )
        self.conv3d_2_3 = nn.Conv3d(
            in_channels=16, out_channels=32, kernel_size=1, stride=1
        )
        self.batch_norm_2_1 = nn.BatchNorm3d(24)
        self.batch_norm_2_2 = nn.BatchNorm3d(32)
        self.batch_norm_2_3 = nn.BatchNorm3d(32)

        # Third convolution block
        self.conv3d_3_1 = nn.Conv3d(
            in_channels=32, out_channels=40, kernel_size=3, stride=1, padding=1
        )
        self.conv3d_3_2 = nn.Conv3d(
            in_channels=40, out_channels=48, kernel_size=3, stride=1, padding=1
        )
        self.conv3d_3_3 = nn.Conv3d(
            in_channels=32, out_channels=48, kernel_size=1, stride=1
        )
        self.batch_norm_3_1 = nn.BatchNorm3d(40)
        self.batch_norm_3_2 = nn.BatchNorm3d(48)
        self.batch_norm_3_3 = nn.BatchNorm3d(48)

        # Fourth convolution block
        self.conv3d_4_1 = nn.Conv3d(
            in_channels=48, out_channels=56, kernel_size=3, stride=1, padding=1
        )
        self.conv3d_4_2 = nn.Conv3d(
            in_channels=56, out_channels=64, kernel_size=3, stride=1, padding=1
        )
        self.conv3d_4_3 = nn.Conv3d(
            in_channels=48, out_channels=64, kernel_size=1, stride=1
        )
        self.batch_norm_4_1 = nn.BatchNorm3d(56)
        self.batch_norm_4_2 = nn.BatchNorm3d(64)
        self.batch_norm_4_3 = nn.BatchNorm3d(64)

        self.pool = nn.MaxPool3d(2)
        self.global_pool = nn.AdaptiveAvgPool3d((1, 1, 1))
        self.dropout = nn.Dropout(0.3)
        self.fc1 = nn.Linear(64, 32)
        self.num_classes = num_classes

    @Model.num_classes.setter  # type: ignore[attr-defined]
    def num_classes(self, v: int) -> None:
        self.fc2 = nn.Linear(32, v)
        self._num_classes = v

    def on_load_checkpoint(self, checkpoint: Checkpoint) -> None:
        num_classes = checkpoint["state_dict"]["fc2.bias"].shape[0]

        if num_classes != self.num_classes:
            logger.debug(
                f"Resetting number-of-output-classes at `{self.name}` model from "
                f"{self.num_classes} to {num_classes} while loading checkpoint."
            )
        self.num_classes = num_classes

        super().on_load_checkpoint(checkpoint)

    def forward(self, x):
        x = self.normalizer(x)  # type: ignore

        # First convolution block
        _x = x
        x = F.relu(self.batch_norm_1_1(self.conv3d_1_1(x)))
        x = F.relu(self.batch_norm_1_2(self.conv3d_1_2(x)))
        x = (x + F.relu(self.batch_norm_1_3(self.conv3d_1_3(_x)))) / 2
        x = self.pool(x)

        # Second convolution block

        _x = x
        x = F.relu(self.batch_norm_2_1(self.conv3d_2_1(x)))
        x = F.relu(self.batch_norm_2_2(self.conv3d_2_2(x)))
        x = (x + F.relu(self.batch_norm_2_3(self.conv3d_2_3(_x)))) / 2
        x = self.pool(x)

        # Third convolution block

        _x = x
        x = F.relu(self.batch_norm_3_1(self.conv3d_3_1(x)))
        x = F.relu(self.batch_norm_3_2(self.conv3d_3_2(x)))
        x = (x + F.relu(self.batch_norm_3_3(self.conv3d_3_3(_x)))) / 2
        x = self.pool(x)

        # Fourth convolution block

        _x = x
        x = F.relu(self.batch_norm_4_1(self.conv3d_4_1(x)))
        x = F.relu(self.batch_norm_4_2(self.conv3d_4_2(x)))
        x = (x + F.relu(self.batch_norm_4_3(self.conv3d_4_3(_x)))) / 2

        x = self.global_pool(x)
        x = x.view(x.size(0), x.size(1))

        x = F.relu(self.fc1(x))
        x = self.dropout(x)
        return self.fc2(x)

        # x = F.log_softmax(x, dim=1) # 0 is batch size



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""Logistic regression classifier, to be trained from scratch.

It is preset to take 14 inputs (e.g. from the output of a model for
radiological sign classification based on :py:mod:`.data.nih_cxr14`).
"""

from mednet.models.classify.logistic_regression import LogisticRegression

model = LogisticRegression(input_size=14)



# SPDX-FileCopyrightText: Copyright © 2024 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
""":py:mod:`AV-DRIVE database <mednet.data.segment.avdrive>` (default split)."""

import importlib.resources

from mednet.data.segment.avdrive import DataModule

datamodule = DataModule(
    importlib.resources.files(__package__ or __name__.rsplit(".", 1)[0])
    / "default.json"
)



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""Data loading code."""

import numpy.typing
import PIL.Image


def remove_black_borders(
    img: PIL.Image.Image,
    threshold: int = 0,
) -> tuple[PIL.Image.Image, numpy.typing.NDArray[numpy.bool_]]:
    """Remove black borders of CXR.

    Parameters
    ----------
    img
        A PIL image.
    threshold
        Threshold value from which borders are considered black.
        Defaults to 0.

    Returns
    -------
        A PIL image with black borders removed, and the mask used to remove the
        black borders from the image, that can be subsequently used to process
        other related image information (e.g. annotation masks).
    """

    img_array = numpy.asarray(img)

    if len(img_array.shape) == 2:  # single channel
        mask = numpy.asarray(img_array) > threshold
        return PIL.Image.fromarray(
            img_array[numpy.ix_(mask.any(1), mask.any(0))],
        ), mask

    if len(img_array.shape) == 3 and img_array.shape[2] == 3:
        r_mask = img_array[:, :, 0] > threshold
        g_mask = img_array[:, :, 1] > threshold
        b_mask = img_array[:, :, 2] > threshold

        mask = r_mask | g_mask | b_mask
        return PIL.Image.fromarray(
            img_array[numpy.ix_(mask.any(1), mask.any(0))],
        ), mask

    raise NotImplementedError



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""Shenzhen DataModule for computer-aided diagnosis.

The standard digital image database for Tuberculosis was created by the
National Library of Medicine, Maryland, USA in collaboration with Shenzhen No.3
People’s Hospital, Guangdong Medical College, Shenzhen, China. The Chest X-rays
are from out-patient clinics, and were captured as part of the daily routine
using Philips DR Digital Diagnose systems.

* Database reference: :cite:p:`jaeger_two_2014`

.. important:: **Raw data organization**

    The Shenzhen_ base datadir, which you should configure following the
    :ref:`mednet.setup` instructions, must contain at this subdirectory:

    - ``CXR_png/`` (directory containing the CXR images)

Data specifications:

* Raw data input (on disk):

  * PNG 8-bit RGB images issued from digital radiography machines (grayscale,
    but encoded as RGB images with "inverted" grayscale scale requiring special
    treatment).
  * Original resolution: variable width and height of 3000 x 3000 pixels or
    less
  * Samples: 662 images and associated labels

* Output image:

  * Transforms:

    * Load raw data with :py:mod:`PIL` with auto-conversion to grayscale
    * Remove (completely) black borders
    * Convert to torch tensor

  * Final specifications:

    * Grayscale, encoded as a single plane tensor, 32-bit floats,
      square with varying resolutions, depending on the input image
    * Labels: 0 (healthy), 1 (active tuberculosis)

This module contains the base declaration of common data modules and raw-data
loaders for this database. All configured splits inherit from this definition.
"""

import importlib.resources.abc
import os
import pathlib
import typing

import PIL.Image
import torch
from torchvision import tv_tensors
from torchvision.transforms.v2.functional import to_dtype, to_image

from ...utils.rc import load_rc
from ..datamodule import CachingDataModule
from ..image_utils import remove_black_borders
from ..split import JSONDatabaseSplit
from ..typing import RawDataLoader as BaseDataLoader
from ..typing import Sample

DATABASE_SLUG = __name__.rsplit(".", 1)[-1]
"""Pythonic name of this database."""

CONFIGURATION_KEY_DATADIR = "datadir." + DATABASE_SLUG
"""Key to search for in the configuration file for the root directory of this
database."""


class RawDataLoader(BaseDataLoader):
    """A specialized raw-data-loader for the Shenzhen dataset.

    Parameters
    ----------
    config_variable
        Key to search for in the configuration file for the root directory of
        this database.
    """

    datadir: pathlib.Path
    """This variable contains the base directory where the database raw data is
    stored."""

    # config_variable: required so this loader can be used for the Indian
    # database as well.
    def __init__(self, config_variable: str = CONFIGURATION_KEY_DATADIR):
        self.datadir = pathlib.Path(
            load_rc().get(config_variable, os.path.realpath(os.curdir)),
        )

    def sample(self, sample: typing.Any) -> Sample:
        """Load a single image sample from the disk.

        Parameters
        ----------
        sample
            A tuple containing the path suffix, within the dataset root folder,
            where to find the image to be loaded, and an integer, representing
            the sample target.

        Returns
        -------
            The sample representation.
        """

        # N.B.: Image.convert("L") is required to normalize grayscale back to
        # normal (instead of inverted).
        image = PIL.Image.open(self.datadir / sample[0]).convert("L")
        image, _ = remove_black_borders(image)
        image = to_dtype(to_image(image), torch.float32, scale=True)
        image = tv_tensors.Image(image)

        # use the code below to view generated images
        # from torchvision.transforms.v2.functional import to_pil_image
        # to_pil_image(tensor).show()
        # __import__("pdb").set_trace()

        return dict(image=image, target=self.target(sample), name=sample[0])

    def target(self, sample: typing.Any) -> torch.Tensor:
        """Load only sample target from its raw representation.

        Parameters
        ----------
        sample
            A tuple containing the path suffix, within the dataset root folder,
            where to find the image to be loaded, and an integer, representing
            the sample target.

        Returns
        -------
            The label corresponding to the specified sample, encapsulated as a
            1D torch float tensor.
        """

        return torch.FloatTensor([sample[1]])


class DataModule(CachingDataModule):
    """Shenzhen DataModule for computer-aided diagnosis.

    Parameters
    ----------
    split_path
        Path or traversable (resource) with the JSON split description to load.
    """

    def __init__(self, split_path: pathlib.Path | importlib.resources.abc.Traversable):
        super().__init__(
            database_split=JSONDatabaseSplit(split_path),
            raw_data_loader=RawDataLoader(),
            database_name=DATABASE_SLUG,
            split_name=split_path.name.rsplit(".", 2)[0],
            task="classification",
            num_classes=1,
        )



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""Tests for VISCERAL dataset."""

import pytest
from click.testing import CliRunner


def id_function(val):
    if isinstance(val, dict):
        return str(val)
    return repr(val)


@pytest.mark.parametrize(
    "split,lenghts",
    [
        ("default", dict(train=241, validation=69, test=35)),
    ],
    ids=id_function,  # just changes how pytest prints it
)
def test_protocol_consistency(
    database_checkers,
    split: str,
    lenghts: dict[str, int],
):
    from mednet.data.split import make_split

    database_checkers.check_split(
        make_split("mednet.config.classify.data.visceral", f"{split}.json"),
        lengths=lenghts,
        prefixes=("16/10000"),
        possible_labels=(0, 1),
    )


@pytest.mark.skip_if_rc_var_not_set("datadir.visceral")
def test_database_check():
    from mednet.scripts.database import check

    runner = CliRunner()
    result = runner.invoke(check, ["visceral"])
    assert (
        result.exit_code == 0
    ), f"Exit code {result.exit_code} != 0 -- Output:\n{result.output}"



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""Simple CNN network model from :cite:p:`pasa_efficient_2019`."""

import logging
import typing

import torch
import torch.nn
import torch.nn.functional as F  # noqa: N812
import torch.optim.optimizer
import torch.utils.data

from ...data.typing import TransformSequence
from ..typing import Checkpoint
from .model import Model

logger = logging.getLogger(__name__)


class Pasa(Model):
    """Simple CNN network model from :cite:p:`pasa_efficient_2019`.

    This network has a linear output.  You should use losses with ``WithLogit``
    instead of cross-entropy versions when training.

    Parameters
    ----------
    loss_type
        The loss to be used for training and evaluation.

        .. warning::

           The loss should be set to always return batch averages (as opposed
           to the batch sum), as our logging system expects it so.
    loss_arguments
        Arguments to the loss.
    optimizer_type
        The type of optimizer to use for training.
    optimizer_arguments
        Arguments to the optimizer after ``params``.
    scheduler_type
        The type of scheduler to use for training.
    scheduler_arguments
        Arguments to the scheduler after ``params``.
    model_transforms
        An optional sequence of torch modules containing transforms to be
        applied on the input **before** it is fed into the network.
    augmentation_transforms
        An optional sequence of torch modules containing transforms to be
        applied on the input **before** it is fed into the network.
    num_classes
        Number of outputs (classes) for this model.
    """

    def __init__(
        self,
        loss_type: type[torch.nn.Module] = torch.nn.BCEWithLogitsLoss,
        loss_arguments: dict[str, typing.Any] | None = None,
        optimizer_type: type[torch.optim.Optimizer] = torch.optim.Adam,
        optimizer_arguments: dict[str, typing.Any] | None = None,
        scheduler_type: type[torch.optim.lr_scheduler.LRScheduler] | None = None,
        scheduler_arguments: dict[str, typing.Any] | None = None,
        model_transforms: TransformSequence | None = None,
        augmentation_transforms: TransformSequence | None = None,
        num_classes: int = 1,
    ):
        super().__init__(
            name="pasa",
            loss_type=loss_type,
            loss_arguments=loss_arguments,
            optimizer_type=optimizer_type,
            optimizer_arguments=optimizer_arguments,
            scheduler_type=scheduler_type,
            scheduler_arguments=scheduler_arguments,
            model_transforms=model_transforms,
            augmentation_transforms=augmentation_transforms,
            num_classes=num_classes,
        )

        # First convolution block
        self.fc1 = torch.nn.Conv2d(1, 4, (3, 3), (2, 2), (1, 1))
        self.fc2 = torch.nn.Conv2d(4, 16, (3, 3), (2, 2), (1, 1))
        self.fc3 = torch.nn.Conv2d(1, 16, (1, 1), (4, 4))

        self.batchNorm2d_4 = torch.nn.BatchNorm2d(4)
        self.batchNorm2d_16 = torch.nn.BatchNorm2d(16)
        self.batchNorm2d_16_2 = torch.nn.BatchNorm2d(16)

        # Second convolution block
        self.fc4 = torch.nn.Conv2d(16, 24, (3, 3), (1, 1), (1, 1))
        self.fc5 = torch.nn.Conv2d(24, 32, (3, 3), (1, 1), (1, 1))
        self.fc6 = torch.nn.Conv2d(
            16,
            32,
            (1, 1),
            (1, 1),
        )  # Original stride (2, 2)

        self.batchNorm2d_24 = torch.nn.BatchNorm2d(24)
        self.batchNorm2d_32 = torch.nn.BatchNorm2d(32)
        self.batchNorm2d_32_2 = torch.nn.BatchNorm2d(32)

        # Third convolution block
        self.fc7 = torch.nn.Conv2d(32, 40, (3, 3), (1, 1), (1, 1))
        self.fc8 = torch.nn.Conv2d(40, 48, (3, 3), (1, 1), (1, 1))
        self.fc9 = torch.nn.Conv2d(
            32,
            48,
            (1, 1),
            (1, 1),
        )  # Original stride (2, 2)

        self.batchNorm2d_40 = torch.nn.BatchNorm2d(40)
        self.batchNorm2d_48 = torch.nn.BatchNorm2d(48)
        self.batchNorm2d_48_2 = torch.nn.BatchNorm2d(48)

        # Fourth convolution block
        self.fc10 = torch.nn.Conv2d(48, 56, (3, 3), (1, 1), (1, 1))
        self.fc11 = torch.nn.Conv2d(56, 64, (3, 3), (1, 1), (1, 1))
        self.fc12 = torch.nn.Conv2d(
            48,
            64,
            (1, 1),
            (1, 1),
        )  # Original stride (2, 2)

        self.batchNorm2d_56 = torch.nn.BatchNorm2d(56)
        self.batchNorm2d_64 = torch.nn.BatchNorm2d(64)
        self.batchNorm2d_64_2 = torch.nn.BatchNorm2d(64)

        # Fifth convolution block
        self.fc13 = torch.nn.Conv2d(64, 72, (3, 3), (1, 1), (1, 1))
        self.fc14 = torch.nn.Conv2d(72, 80, (3, 3), (1, 1), (1, 1))
        self.fc15 = torch.nn.Conv2d(
            64,
            80,
            (1, 1),
            (1, 1),
        )  # Original stride (2, 2)

        self.batchNorm2d_72 = torch.nn.BatchNorm2d(72)
        self.batchNorm2d_80 = torch.nn.BatchNorm2d(80)
        self.batchNorm2d_80_2 = torch.nn.BatchNorm2d(80)

        self.pool2d = torch.nn.MaxPool2d(
            (3, 3),
            (2, 2),
        )  # Pool after conv. block

        self.dense = torch.nn.Linear(80, self.num_classes)  # Fully connected layer

    @Model.num_classes.setter  # type: ignore[attr-defined]
    def num_classes(self, v: int) -> None:
        if self.num_classes != v:
            logger.info(
                f"Resetting `{self.name}` output classifier layer weights due "
                f"to change in output size ({self.num_classes} -> {v})"
            )
            self.dense = torch.nn.Linear(80, v)  # Fully connected layer
            self._num_classes = v

    def on_load_checkpoint(self, checkpoint: Checkpoint) -> None:
        # reset number of output classes if need be
        self.num_classes = checkpoint["state_dict"]["dense.bias"].shape[0]

        # perform routine checkpoint loading
        super().on_load_checkpoint(checkpoint)

    def forward(self, x):
        x = self.normalizer(x)  # type: ignore

        # First convolution block
        _x = x
        x = F.relu(self.batchNorm2d_4(self.fc1(x)))  # 1st convolution
        x = F.relu(self.batchNorm2d_16(self.fc2(x)))  # 2nd convolution
        x = (x + F.relu(self.batchNorm2d_16_2(self.fc3(_x)))) / 2  # Parallel
        x = self.pool2d(x)  # Pooling

        # Second convolution block
        _x = x
        x = F.relu(self.batchNorm2d_24(self.fc4(x)))  # 1st convolution
        x = F.relu(self.batchNorm2d_32(self.fc5(x)))  # 2nd convolution
        x = (x + F.relu(self.batchNorm2d_32_2(self.fc6(_x)))) / 2  # Parallel
        x = self.pool2d(x)  # Pooling

        # Third convolution block
        _x = x
        x = F.relu(self.batchNorm2d_40(self.fc7(x)))  # 1st convolution
        x = F.relu(self.batchNorm2d_48(self.fc8(x)))  # 2nd convolution
        x = (x + F.relu(self.batchNorm2d_48_2(self.fc9(_x)))) / 2  # Parallel
        x = self.pool2d(x)  # Pooling

        # Fourth convolution block
        _x = x
        x = F.relu(self.batchNorm2d_56(self.fc10(x)))  # 1st convolution
        x = F.relu(self.batchNorm2d_64(self.fc11(x)))  # 2nd convolution
        x = (x + F.relu(self.batchNorm2d_64_2(self.fc12(_x)))) / 2  # Parallel
        x = self.pool2d(x)  # Pooling

        # Fifth convolution block
        _x = x
        x = F.relu(self.batchNorm2d_72(self.fc13(x)))  # 1st convolution
        x = F.relu(self.batchNorm2d_80(self.fc14(x)))  # 2nd convolution
        x = (x + F.relu(self.batchNorm2d_80_2(self.fc15(_x)))) / 2  # Parallel
        # no pooling

        # Global average pooling
        x = torch.mean(x.view(x.size(0), x.size(1), -1), dim=2)

        # Dense layer
        return self.dense(x)

        # x = F.log_softmax(x, dim=1) # 0 is batch size



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import unittest

import mednet.config.classify.models.pasa as pasa_config
from mednet.utils.summary import summary


class Tester(unittest.TestCase):
    """Unit test for model architectures."""

    def test_summary_driu(self):
        model = pasa_config.model
        s, param = summary(model)
        self.assertIsInstance(s, str)
        self.assertIsInstance(param, int)



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""Engine and functions for saliency map generation."""

import logging
import pathlib
import typing

import lightning.pytorch
import numpy
import torch
import torch.nn
import tqdm

from ....engine.device import DeviceManager
from ....models.classify.typing import SaliencyMapAlgorithm

logger = logging.getLogger(__name__)


def _create_saliency_map_callable(
    algo_type: SaliencyMapAlgorithm,
    model: torch.nn.Module,
    target_layers: list[torch.nn.Module] | None,
):
    """Create a class activation map (CAM) instance for a given model.

    Parameters
    ----------
    algo_type
        The algorithm to use for saliency map estimation.
    model
        Neural network model (e.g. pasa).
    target_layers
        The target layers to compute CAM for.

    Returns
    -------
        A class activation map (CAM) instance for the given model.
    """

    import pytorch_grad_cam

    match algo_type:
        case "gradcam":
            return pytorch_grad_cam.GradCAM(
                model=model,
                target_layers=target_layers,
            )
        case "scorecam":
            return pytorch_grad_cam.ScoreCAM(
                model=model,
                target_layers=target_layers,
            )
        case "fullgrad":
            return pytorch_grad_cam.FullGrad(
                model=model,
                target_layers=target_layers,
            )
        case "randomcam":
            return pytorch_grad_cam.RandomCAM(
                model=model,
                target_layers=target_layers,
            )
        case "hirescam":
            return pytorch_grad_cam.HiResCAM(
                model=model,
                target_layers=target_layers,
            )
        case "gradcamelementwise":
            return pytorch_grad_cam.GradCAMElementWise(
                model=model,
                target_layers=target_layers,
            )
        case "gradcam++" | "gradcamplusplus":
            return pytorch_grad_cam.GradCAMPlusPlus(
                model=model,
                target_layers=target_layers,
            )
        case "xgradcam":
            return pytorch_grad_cam.XGradCAM(
                model=model,
                target_layers=target_layers,
            )
        case "ablationcam":
            assert (
                target_layers is not None
            ), "AblationCAM cannot have target_layers=None"
            return pytorch_grad_cam.AblationCAM(
                model=model,
                target_layers=target_layers,
            )
        case "eigencam":
            return pytorch_grad_cam.EigenCAM(
                model=model,
                target_layers=target_layers,
            )
        case "eigengradcam":
            return pytorch_grad_cam.EigenGradCAM(
                model=model,
                target_layers=target_layers,
            )
        case "layercam":
            return pytorch_grad_cam.LayerCAM(
                model=model,
                target_layers=target_layers,
            )
        case _:
            raise ValueError(
                f"Saliency map algorithm `{algo_type}` is not currently " f"supported.",
            )


def _save_saliency_map(
    output_folder: pathlib.Path,
    name: str,
    saliency_map: torch.Tensor,
) -> None:
    """Save a saliency map to permanent storage (disk).

    Helper function to save a saliency map to disk.

    Parameters
    ----------
    output_folder
        Directory in which the resulting saliency maps will be saved.
    name
        Name of the saved file.
    saliency_map
        A real-valued saliency-map that conveys regions used for
        classification in the original sample.
    """

    n = pathlib.Path(name)
    (output_folder / n.parent).mkdir(parents=True, exist_ok=True)
    numpy.save(output_folder / n.with_suffix(".npy"), saliency_map[0])


def run(
    model: lightning.pytorch.LightningModule,
    datamodule: lightning.pytorch.LightningDataModule,
    device_manager: DeviceManager,
    saliency_map_algorithm: SaliencyMapAlgorithm,
    target_class: typing.Literal["highest", "all"],
    positive_only: bool,
    output_folder: pathlib.Path,
    only_dataset: str | None,
) -> None:
    """Apply saliency mapping techniques on input CXR, outputs pickled saliency
    maps directly to disk.

    Parameters
    ----------
    model
        Neural network model (e.g. pasa).
    datamodule
        The lightning DataModule to iterate on.
    device_manager
        An internal device representation, to be used for training and
        validation.  This representation can be converted into a pytorch device
        or a lightning accelerator setup.
    saliency_map_algorithm
        The algorithm to use for saliency map estimation.
    target_class
        (Use only with multi-label models) Which class to target for CAM
        calculation. Can be either set to "all" or "highest". "highest" is
        default, which means only saliency maps for the class with the highest
        activation will be generated.
    positive_only
        If set, saliency maps will only be generated for positive samples (ie.
        label == 1 in a binary classification task).  This option is ignored on
        a multi-class output model.
    output_folder
        Where to save all the saliency maps (this path should exist before
        this function is called).
    only_dataset
        If set, will only run this code for the named dataset on the provided
        datamodule, skipping any other datasets.
    """

    from pytorch_grad_cam.utils.model_targets import ClassifierOutputTarget

    from ....models.classify.densenet import Densenet
    from ....models.classify.pasa import Pasa

    if isinstance(model, Pasa):
        if saliency_map_algorithm == "fullgrad":
            raise ValueError(
                "Fullgrad saliency map algorithm is not supported for the "
                "Pasa model.",
            )
        target_layers = [model.fc14]  # Last non-1x1 Conv2d layer
    elif isinstance(model, Densenet):
        target_layers = [
            model.model.features.denseblock4.denselayer16.conv2,  # type: ignore
        ]
    else:
        raise TypeError(f"Model of type `{type(model)}` is not yet supported.")

    # prepares model for evaluation, cast to target device
    device = device_manager.torch_device()
    model = model.to(device)
    model.eval()

    saliency_map_callable = _create_saliency_map_callable(
        saliency_map_algorithm,
        model,
        target_layers,  # type: ignore
    )

    for k, v in datamodule.predict_dataloader().items():
        if only_dataset is not None and k != only_dataset:
            logger.warning(
                f"Skipping processing for dataset `{k}` following user request..."
            )
            continue

        logger.info(
            f"Generating saliency maps for dataset `{k}` via "
            f"`{saliency_map_algorithm}`...",
        )

        for sample in tqdm.tqdm(v, desc="samples", leave=False, disable=None):
            name = sample["name"][0]
            target = sample["target"].item()
            image = sample["image"].to(
                device=device,
                non_blocking=torch.cuda.is_available(),
            )

            # in binary classification systems, negative targets may be skipped
            if positive_only and (model.num_classes == 1) and (target == 0):
                continue

            # chooses target outputs to generate saliency maps for
            if model.num_classes > 1:
                if target_class == "all":
                    # just blindly generate saliency maps for all outputs
                    # - make one directory for every target output and lay
                    # images there like in the original dataset.
                    for output_num in range(model.num_classes):
                        use_folder = output_folder / str(output_num)
                        saliency_map = saliency_map_callable(
                            input_tensor=image,
                            targets=[ClassifierOutputTarget(output_num)],  # type: ignore
                        )
                        _save_saliency_map(use_folder, name, saliency_map)  # type: ignore

                else:
                    # pytorch-grad-cam will figure out the output with the
                    # highest value and produce a saliency map for it - we
                    # will save it to disk.
                    use_folder = output_folder / "highest-output"
                    saliency_map = saliency_map_callable(
                        input_tensor=image,
                        # setting `targets=None` will set target to the
                        # maximum output index using
                        # ClassifierOutputTarget(max_output_index)
                        targets=None,  # type: ignore
                    )
                    _save_saliency_map(use_folder, name, saliency_map)  # type: ignore
            else:
                # binary classification model with a single output - just
                # lay all cams uniformily like the original dataset
                saliency_map = saliency_map_callable(
                    input_tensor=image,
                    targets=[
                        ClassifierOutputTarget(0),  # type: ignore
                    ],
                )
                _save_saliency_map(output_folder, name, saliency_map)  # type: ignore



# SPDX-FileCopyrightText: Copyright © 2024 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""RIM-ONE r3 (training set) for cup segmentation.

The dataset contains 159 stereo eye fundus images with a resolution of 2144 x
1424. The right part of the stereo image is disregarded. Two sets of
ground-truths for optic disc and optic cup are available. The first set is
commonly used for training and testing. The second set acts as a “human”
baseline.  A third set, composed of annotation averages may also be used for
training and evaluation purposes.

* Reference: :cite:p:`fumero_interactive_2015`
* Original resolution (height x width): 1424 x 1072
* Split reference: :cite:p:`maninis_deep_2016`
* Protocols ``optic-disc-exp1``, ``optic-cup-exp1``, ``optic-disc-exp2``,
  ``optic-cup-exp2``, ``optic-disc-avg`` and ``optic-cup-avg``
* Training: 99
* Test: 60

This module contains the base declaration of common data modules and raw-data
loaders for this database. All configured splits inherit from this definition.
"""

import importlib.resources
import importlib.resources.abc
import os
import pathlib
import typing

import PIL.Image
import torch
from torchvision import tv_tensors
from torchvision.transforms.v2.functional import to_dtype, to_image

from ...models.transforms import crop_image_to_mask
from ...utils.rc import load_rc
from ..datamodule import CachingDataModule
from ..split import JSONDatabaseSplit
from ..typing import RawDataLoader as BaseDataLoader
from ..typing import Sample

DATABASE_SLUG = __name__.rsplit(".", 1)[-1]
"""Pythonic name to refer to this database."""

CONFIGURATION_KEY_DATADIR = "datadir." + DATABASE_SLUG
"""Key to search for in the configuration file for the root directory of this
database."""

_CROP_COORDINATES = (0, 0, 1072, 1424)
"""Crop RIM-ONE r3 image to avoid (right-side) stereo pair.

* Coordinates are (torchvision): top = 0, left = 0, height = 1424, width = 1072
* Coordinates are (PIL): left = 0, upper = 0, right = 1072, lower = 1424
"""


class RawDataLoader(BaseDataLoader):
    """A specialized raw-data-loader for the rimoner3 dataset."""

    datadir: pathlib.Path
    """This variable contains the base directory where the database raw data is
    stored."""

    def __init__(self):
        self.datadir = pathlib.Path(
            load_rc().get(CONFIGURATION_KEY_DATADIR, os.path.realpath(os.curdir))
        )

    def sample(self, sample: typing.Any) -> Sample:
        """Load a single image sample from the disk.

        Parameters
        ----------
        sample
            A tuple containing the path suffix, within the dataset root folder,
            where to find the image to be loaded, and an integer, representing the
            sample label.

        Returns
        -------
            The sample representation.
        """

        image = (
            PIL.Image.open(self.datadir / sample[0])
            .convert(mode="RGB")
            .crop(_CROP_COORDINATES)
        )
        image = to_dtype(to_image(image), torch.float32, scale=True)

        target = self.target(sample)

        mask_path = (
            importlib.resources.files(__package__) / "masks" / DATABASE_SLUG / sample[2]
        )
        with importlib.resources.as_file(mask_path) as path:
            mask = (
                PIL.Image.open(path)
                .convert(mode="1", dither=None)
                .crop(_CROP_COORDINATES)
            )
            mask = to_dtype(to_image(mask), torch.float32, scale=True)

        image = tv_tensors.Image(crop_image_to_mask(image, mask))
        target = tv_tensors.Mask(crop_image_to_mask(target, mask))
        mask = tv_tensors.Mask(mask)

        return dict(image=image, target=target, mask=mask, name=sample[0])

    def target(self, sample: typing.Any) -> torch.Tensor:
        """Load only sample target from its raw representation.

        Parameters
        ----------
        sample
            A tuple containing the path suffix, within the dataset root folder,
            where to find the image to be loaded, and an integer, representing
            the sample target.

        Returns
        -------
            The label corresponding to the specified sample, encapsulated as a
            torch float tensor.
        """

        target = (
            PIL.Image.open(self.datadir / sample[1])
            .convert(mode="1", dither=None)
            .crop(_CROP_COORDINATES)
        )
        return to_dtype(to_image(target), torch.float32, scale=True)


class DataModule(CachingDataModule):
    """RIM-ONE r3 (training set) for cup segmentation.

    Parameters
    ----------
    split_path
        Path or traversable (resource) with the JSON split description to load.
    """

    def __init__(self, split_path: pathlib.Path | importlib.resources.abc.Traversable):
        super().__init__(
            database_split=JSONDatabaseSplit(split_path),
            raw_data_loader=RawDataLoader(),
            database_name=DATABASE_SLUG,
            split_name=split_path.name.rsplit(".", 2)[0],
            task="segmentation",
            num_classes=1,
        )



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import pathlib
import typing

import numpy
import pytest
import torch

from mednet.data.split import JSONDatabaseSplit
from mednet.data.typing import DatabaseSplit


@pytest.fixture
def datadir(request) -> pathlib.Path:
    """Return the directory in which the test is sitting. Check the pytest
    documentation for more information.

    Parameters
    ----------
    request
        Information of the requesting test function.

    Returns
    -------
    pathlib.Path
        The directory in which the test is sitting.
    """

    return pathlib.Path(request.module.__file__).parents[0] / "data"


def pytest_configure(config):
    """This function is run once for pytest setup.

    Parameters
    ----------
    config
        Configuration values. Check the pytest documentation for more
        information.
    """

    config.addinivalue_line(
        "markers",
        "skip_if_rc_var_not_set(name): this mark skips the test if a certain "
        "~/.config/mednet.toml variable is not set",
    )

    config.addinivalue_line("markers", "slow: this mark indicates slow tests")


def pytest_runtest_setup(item):
    """This function is run for every test candidate in this directory.

    The test is run if this function returns ``None``.  To skip a test,
    call ``pytest.skip()``, specifying a reason.

    Parameters
    ----------
    item
        A test invocation item. Check the pytest documentation for more
        information.
    """

    from mednet.utils.rc import load_rc

    rc = load_rc()

    # iterates over all markers for the item being examined, get the first
    # argument and accumulate these names
    rc_names = [
        mark.args[0] for mark in item.iter_markers(name="skip_if_rc_var_not_set")
    ]

    # checks all names mentioned are set in ~/.config/mednet.toml, otherwise,
    # skip the test
    if rc_names:
        missing = [k for k in rc_names if rc.get(k) is None]
        if any(missing):
            pytest.skip(
                f"Test skipped because {', '.join(missing)} is **not** "
                f"set in ~/.config/mednet.toml",
            )


def rc_variable_set(name):
    from mednet.utils.rc import load_rc

    rc = load_rc()
    pytest.mark.skipif(
        name not in rc,
        reason=f"RC variable '{name}' is not set",
    )


@pytest.fixture(scope="session")
def session_tmp_path(tmp_path_factory):
    return tmp_path_factory.mktemp("test-cli")


class DatabaseCheckers:
    """Helpers for database tests."""

    @staticmethod
    def check_split(
        split: DatabaseSplit,
        lengths: dict[str, int],
        prefixes: typing.Sequence[str],
        possible_labels: typing.Sequence[int],
    ):
        """Run a simple consistency check on the data split.

        Parameters
        ----------
        split
            An instance of DatabaseSplit.
        lengths
            A dictionary that contains keys matching those of the split (this
            will be checked).  The values of the dictionary should correspond
            to the sizes of each of the datasets in the split.
        prefixes
            Each file named in a split should start with at least one of these
            prefixes.
        possible_labels
            These are the list of possible labels contained in any split.
        """

        assert len(split) == len(lengths)

        for k in lengths.keys():
            # dataset must have been declared
            assert k in split

            assert len(split[k]) == lengths[k]
            for s in split[k]:
                # check filename prefixes match, if prefixes were passed
                if prefixes:
                    assert any([s[0].startswith(k) for k in prefixes]), (
                        f"Sample with name {s[0]} does not start with any of the "
                        f"prefixes in {prefixes}"
                    )

                # check if labels match, if labels were passed
                if possible_labels:
                    if isinstance(s[1], list):
                        assert all([k in possible_labels for k in s[1]])
                    else:
                        assert s[1] in possible_labels

    @staticmethod
    def check_loaded_batch(
        batch,
        batch_size: int,
        color_planes: int,
        prefixes: typing.Sequence[str],
        possible_labels: typing.Sequence[int],
        expected_num_labels: int,
        expected_meta_size: int,
        expected_image_shape: tuple[int, ...] | None = None,
    ):
        """Check the consistency of an individual (loaded) batch.

        Parameters
        ----------
        batch
            The loaded batch to be checked.
        batch_size
            The mini-batch size.
        color_planes
            The number of color planes in the images.
        prefixes
            Each file named in a split should start with at least one of these
            prefixes.
        possible_labels
            These are the list of possible labels contained in any split.
        expected_num_labels
            The expected number of labels each sample should have.
        expected_meta_size
            The expected number of elements on the meta-information dictionary
            on every sample.
        expected_image_shape
            The expected shape of the image (num_channels, width, height).
        """

        assert isinstance(batch, dict)
        assert len(batch) == (expected_meta_size + 1)  # account for image

        assert "image" in batch
        assert isinstance(batch["image"], torch.Tensor)
        assert batch["image"].shape[0] == batch_size  # mini-batch size
        assert batch["image"].shape[1] == color_planes

        if expected_image_shape:
            assert all(
                [data.shape == expected_image_shape for data in batch["image"]],
            )

        assert "target" in batch

        if possible_labels:
            assert all([k in possible_labels for k in batch["target"].ravel()])

        if expected_num_labels:
            assert len(batch["target"]) == expected_num_labels

        assert "name" in batch
        if prefixes:
            assert all(
                [any([k.startswith(j) for j in prefixes]) for k in batch["name"]],
            )

        # use the code below to view generated images
        # from torchvision.transforms.v2.functional import to_pil_image
        # to_pil_image(batch["image"]).show()
        # __import__("pdb").set_trace()

    @staticmethod
    def check_image_quality(
        datamodule,
        reference_histogram_file,
        compare_type="equal",
        pearson_coeff_threshold=0.005,
    ):
        ref_histogram_splits = JSONDatabaseSplit(reference_histogram_file)

        for split_name in ref_histogram_splits:
            raw_samples = datamodule.splits[split_name][0][0]

            # It is not possible to get a sample from a Dataset by name/path,
            # only by index. This creates a dict of sample name to dataset
            # index.
            raw_samples_indices = {}
            for idx, rs in enumerate(raw_samples):
                raw_samples_indices[rs[0]] = idx

            for ref_hist_path, ref_hist_data in ref_histogram_splits[split_name]:
                # Get index in the dataset that will return the data
                # corresponding to the specified sample name
                dataset_sample_index = raw_samples_indices[ref_hist_path]

                image_tensor = datamodule._datasets[split_name][  # noqa: SLF001
                    dataset_sample_index
                ]["image"]

                histogram = []
                for color_channel in image_tensor:
                    color_channel = numpy.multiply(
                        color_channel.numpy(),
                        255,
                    ).astype(int)
                    histogram.extend(
                        numpy.histogram(
                            color_channel,
                            bins=256,
                            range=(0, 256),
                        )[0].tolist(),
                    )

                if compare_type == "statistical":
                    # Compute pearson coefficients between histogram and
                    # reference and check the similarity within a certain
                    # threshold
                    pearson_coeffs = numpy.corrcoef(histogram, ref_hist_data)
                    assert 1 - pearson_coeff_threshold <= pearson_coeffs[0][1] <= 1

                else:
                    assert histogram == ref_hist_data


@pytest.fixture
def database_checkers():
    return DatabaseCheckers



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""Indian database for TB detection (a.k.a. Dataset A/Dataset B).

The Indian collection database has been established to foster research in
computer-aided diagnosis of pulmonary diseases with a special focus on
pulmonary tuberculosis (TB).  This database is also known as the "Database
A/Database B" database.

* Database reference: :cite:p:`noauthor_tbxpredict_2014`
* Split references: :cite:p:`noauthor_tbxpredict_2014` with 20% of train set for the validation
  set

.. important:: **Raw data organization**

    The Indian_ base datadir, which you should configure following the
    :ref:`mednet.setup` instructions, must contain at least these two
    subdirectories:

    - ``DatasetA/`` (directory containing the dataset A images in JPG format)
    - ``DatasetB/`` (directory containing the dataset B images in DICOM format)

Data specifications:

* Raw data input (on disk):

  * JPG RGB 8-bit depth images with "inverted" grayscale scale, with varying
    resolution of at least 1024 x 1024 pixels per sample
  * Samples: 156 images and associated labels

* Output image:  Use the same transforms and specifications as for
  :py:mod:`.classify.shenzhen`

This module contains the base declaration of common data modules and raw-data
loaders for this database. All configured splits inherit from this definition.
"""

import importlib.resources.abc
import pathlib

from ..datamodule import CachingDataModule
from ..split import JSONDatabaseSplit
from .shenzhen import RawDataLoader

DATABASE_SLUG = __name__.rsplit(".", 1)[-1]
"""Pythonic name of this database."""

CONFIGURATION_KEY_DATADIR = "datadir." + DATABASE_SLUG
"""Key to search for in the configuration file for the root directory of this
database."""


class DataModule(CachingDataModule):
    """Indian database for TB detection (a.k.a. Dataset A/Dataset B).
        Names of the JSON files containing the splits to load for montgomery
        and shenzhen databases (in this order).

    Parameters
    ----------
    split_path
        Path or traversable (resource) with the JSON split description to load.
    """

    def __init__(self, split_path: pathlib.Path | importlib.resources.abc.Traversable):
        super().__init__(
            database_split=JSONDatabaseSplit(split_path),
            raw_data_loader=RawDataLoader(config_variable=CONFIGURATION_KEY_DATADIR),
            database_name=DATABASE_SLUG,
            split_name=split_path.name.rsplit(".", 2)[0],
            task="classification",
            num_classes=1,
        )



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""DenseNet_, to be trained from scratch.

This configuration contains a version of DenseNet_ (c.f. `TorchVision's
page <densenet_pytorch_>`), modified for a variable number of outputs
(defaults to 1).
"""

import torch.nn
import torch.optim
import torchvision.transforms
import torchvision.transforms.v2

import mednet.models.classify.densenet
import mednet.models.transforms

model = mednet.models.classify.densenet.Densenet(
    loss_type=torch.nn.BCEWithLogitsLoss,
    optimizer_type=torch.optim.Adam,
    optimizer_arguments=dict(lr=0.0001),
    pretrained=False,
    dropout=0.1,
    model_transforms=[
        mednet.models.transforms.SquareCenterPad(),
        torchvision.transforms.v2.Resize(512, antialias=True),
        torchvision.transforms.v2.RGB(),
    ],
)



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""Engine and functions for visualization of saliency maps."""

import logging
import pathlib
import typing

import lightning.pytorch
import numpy
import numpy.typing
import PIL.Image
import PIL.ImageColor
import PIL.ImageDraw
import torchvision.transforms.v2.functional
from torchvision import tv_tensors
from tqdm import tqdm

logger = logging.getLogger(__name__)


def _overlay_saliency_map(
    image: PIL.Image.Image,
    saliencies: numpy.typing.NDArray[numpy.double],
    colormap: typing.Literal[  # we accept any "Sequential" colormap from mpl
        "viridis",
        "plasma",
        "inferno",
        "magma",
        "cividis",
        "Greys",
        "Purples",
        "Blues",
        "Greens",
        "Oranges",
        "Reds",
        "YlOrBr",
        "YlOrRd",
        "OrRd",
        "PuRd",
        "RdPu",
        "BuPu",
        "GnBu",
        "PuBu",
        "YlGnBu",
        "PuBuGn",
        "BuGn",
        "YlGn",
    ],
    image_weight: float,
) -> PIL.Image.Image:
    """Create an overlayed represention of the saliency map on the original
    image.

    This is a slightly modified version of the show_cam_on_image implementation in:
    https://github.com/jacobgil/pytorch-grad-cam, but uses matplotlib instead
    of opencv.

    Parameters
    ----------
    image
        The input image that will be overlaid with the saliency map.
    saliencies
        The saliency map that will be overlaid on the (raw) image.
    colormap
        The name of the (matplotlib) colormap to be used.
    image_weight
        The final result is ``image_weight * image + (1-image_weight) *
        saliency_map``.

    Returns
    -------
    PIL.Image.Image
        A modified version of the input ``image`` with the overlaid saliency
        map.
    """
    import matplotlib

    image_array = numpy.array(image, dtype=numpy.float32) / 255.0

    assert image_array.shape[:2] == saliencies.shape, (
        f"The shape of the saliency map ({saliencies.shape}) is different "
        f"from the shape of the input image ({image_array.shape[:2]})."
    )

    assert (
        saliencies.max() <= 1
    ), f"The input saliency map should be in the range [0, 1] (max={saliencies.max()})"

    assert (
        image_weight > 0 and image_weight < 1
    ), f"image_weight should be in the range [0, 1], but got {image_weight}"

    heatmap = matplotlib.colormaps[colormap](saliencies)

    # For pixels where the mask is zero, the original image pixels are being
    # used without a mask.
    result = numpy.where(
        saliencies[..., numpy.newaxis] == 0,
        image_array,
        (image_weight * image_array) + ((1 - image_weight) * heatmap[:, :, :3]),
    )

    return PIL.Image.fromarray((result * 255).astype(numpy.uint8), "RGB")


def _process_sample(
    raw_data: numpy.typing.NDArray[numpy.double],
    saliencies: numpy.typing.NDArray[numpy.double],
    ground_truth: tv_tensors.BoundingBoxes,
) -> PIL.Image.Image:
    """Generate an overlaid representation of the original sample and saliency
    maps.

    Parameters
    ----------
    raw_data
        The raw data representing the input sample that will be overlaid with
        saliency maps and annotations.
    saliencies
        The saliency map recovered from the model, that will be imprinted on
        the raw_data.
    ground_truth
        Ground-truth annotations that may be imprinted on the final image.

    Returns
    -------
    PIL.Image.Image
        An image with the original raw data overlaid with the different
        elements as selected by the user.
    """

    # we need a colour image to eventually overlay a (coloured) saliency map on
    # the top, draw rectangles and other annotations in coulour.  So, we force
    # it right up front.
    retval = torchvision.transforms.v2.functional.to_pil_image(raw_data).convert(
        "RGB",
    )

    retval = _overlay_saliency_map(
        retval,
        saliencies,
        colormap="plasma",
        image_weight=0.5,
    )

    if ground_truth is not None:
        retval = torchvision.transforms.v2.functional.to_pil_image(
            torchvision.utils.draw_bounding_boxes(
                tv_tensors.Image(retval), ground_truth[0], colors="green", width=2
            )
        )

    return retval


def run(
    datamodule: lightning.pytorch.LightningDataModule,
    input_folder: pathlib.Path,
    target_label: int,
    output_folder: pathlib.Path,
    show_groundtruth: bool,
    threshold: float,
):
    """Overlay saliency maps on CXR to output final images with heatmaps.

    Parameters
    ----------
    datamodule
        The Lightning DataModule to iterate on.
    input_folder
        Directory in which the saliency maps are stored for a specific
        visualization type.
    target_label
        The label to target for evaluating interpretability metrics. Samples
        contining any other label are ignored.
    output_folder
        Directory in which the resulting visualizations will be saved.
    show_groundtruth
        If set, imprint ground truth labels over the original image and
        saliency maps.
    threshold : float
        The pixel values above ``threshold`` % of max value are kept in the
        original saliency map.  Everything else is set to zero.  The value
        proposed on :cite:p:`wang_score-cam_2020` is 0.2.  Use this value if unsure.
    """

    for dataset_name, dataset_loader in datamodule.predict_dataloader().items():
        logger.info(
            f"Generating visualizations for samples (target = {target_label}) "
            f"at dataset `{dataset_name}`..."
        )

        for sample in tqdm(
            dataset_loader,
            desc="batches",
            leave=False,
            disable=None,
        ):
            # WARNING: following code assumes a batch size of 1. Will break if
            # not the case.
            name = str(sample["name"][0])
            label = int(sample["target"].item())
            data = sample["image"][0]

            if label != target_label:
                # no visualisation was generated
                continue

            saliencies = numpy.load(
                input_folder / pathlib.Path(name).with_suffix(".npy"),
            )
            saliencies[saliencies < (threshold * saliencies.max())] = 0

            # TODO: This is very specific to the TBX11k system for labelling
            # regions of interest.  We need to abstract from this to support more
            # datasets and other ways to annotate.
            if show_groundtruth:
                ground_truth = sample.get("bboxes", None)
            else:
                ground_truth = None

            # we fully process this entry
            image = _process_sample(
                data,
                saliencies,
                ground_truth,
            )

            # Save image
            output_file_path = output_folder / pathlib.Path(name).with_suffix(
                ".png",
            )
            output_file_path.parent.mkdir(parents=True, exist_ok=True)
            image.save(output_file_path)



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""Common utilities for saliency generation and analysis tasks."""

import typing

import matplotlib.figure
import numpy
import numpy.typing
import tabulate


def extract_statistics(
    data: list[tuple[str, int, float, float, float]],
    index: int,
) -> dict[str, typing.Any]:
    """Extract all meaningful statistics from a reconciled statistics set.

    Parameters
    ----------
    data
        A list of tuples each containing a sample name, target, and values
        produced by completeness and interpretability analysis.
    index
        The index of the tuple contained in ``data`` that should be extracted.

    Returns
    -------
        A dictionary containing the following elements:

            * ``values``: A list of values corresponding to the index on the data
            * ``mean``: The mean of the value listdir
            * ``stdev``: The standard deviation of the value list
            * ``quartiles``: The 25%, 50% (median), and 75% quartile of values
            * ``decreasing_scores``: A list of sample names and labels in
              decreasing value.
    """

    val = numpy.array([k[index] for k in data])
    return dict(
        values=val,
        mean=val.mean(),
        stdev=val.std(ddof=1),  # unbiased estimator
        quartiles={
            25: numpy.percentile(val, 25),  # type: ignore
            50: numpy.median(val),  # type: ignore
            75: numpy.percentile(val, 75),  # type: ignore
        },
        decreasing_scores=[
            (k[0], k[index]) for k in sorted(data, key=lambda x: x[index], reverse=True)
        ],
    )


def make_table(
    results: dict[str, list[typing.Any]],
    indexes: dict[int, str],
    format_: str,
) -> str:
    """Summarize results obtained by interpretability or completeness analysis
    in a table.

    Parameters
    ----------
    results
        The results to be summarized.
    indexes
        A dictionary where keys are indexes in each sample of ``results``, and
        values are a (possibly abbreviated) name to be used in table headers.
    format_
        The table format.

    Returns
    -------
        A table, formatted following ``format_`` and containing the
        various quartile informations for each split and metric.
    """

    headers = ["subset", "samples"]
    for idx, name in indexes.items():
        headers += [
            f"{name}[mean]",
            f"{name}[std]",
            f"{name}[25%]",
            f"{name}[50%]",
            f"{name}[75%]",
        ]

    data = []

    for k, v in results.items():
        samples = [s for s in v if len(s) != 2]
        row = [k, len(samples)]
        for idx in indexes.keys():
            stats = extract_statistics(samples, index=idx)
            row += [
                stats["mean"],
                stats["stdev"],
                stats["quartiles"][25],
                stats["quartiles"][50],
                stats["quartiles"][75],
            ]
        data.append(row)

    return tabulate.tabulate(data, headers, tablefmt=format_, floatfmt=".3f")


def make_histogram(
    name: str,
    values: numpy.typing.NDArray,
    xlim: tuple[float, float] | None = None,
    title: None | str = None,
) -> matplotlib.figure.Figure:
    """Build an histogram of values.

    Parameters
    ----------
    name
        Name of the variable to be histogrammed (will appear in the figure).
    values
        Values to be histogrammed.
    xlim
        A tuple representing the X-axis maximum and minimum to plot. If not
        set, then use the bin boundaries.
    title
        A title to set on the histogram.

    Returns
    -------
        A matplotlib figure containing the histogram.
    """

    from matplotlib import pyplot

    fig, ax = pyplot.subplots(1)
    ax = typing.cast(matplotlib.figure.Axes, ax)
    ax.set_xlabel(name)
    ax.set_ylabel("Frequency")

    if title is not None:
        ax.set_title(title)
    else:
        ax.set_title(f"{name} Frequency Histogram")

    n, bins, _ = ax.hist(values, bins="auto", density=True, alpha=0.7)

    if xlim is not None:
        ax.spines.bottom.set_bounds(*xlim)
    else:
        ax.spines.bottom.set_bounds(bins[0], bins[-1])

    ax.spines.left.set_bounds(0, n.max())
    ax.spines.right.set_visible(False)
    ax.spines.top.set_visible(False)

    ax.grid(linestyle="--", linewidth=1, color="gray", alpha=0.3)

    # draw median and quartiles
    quartile = numpy.percentile(values, [25, 50, 75])
    ax.axvline(
        quartile[0],
        color="green",
        linestyle="--",
        label="Q1",
        alpha=0.5,
    )
    ax.axvline(quartile[1], color="red", label="median", alpha=0.5)
    ax.axvline(
        quartile[2],
        color="green",
        linestyle="--",
        label="Q3",
        alpha=0.5,
    )

    return fig  # type: ignore


def make_plots(
    results: dict[str, list[typing.Any]],
    indexes: dict[int, str],
    xlim: tuple[float, float] | None = None,
) -> list[matplotlib.figure.Figure]:
    """Plot histograms for a particular variable, across all datasets.

    Parameters
    ----------
    results
        The results to be plotted.
    indexes
        A dictionary where keys are indexes in each sample of ``results``, and
        values are a (possibly abbreviated) name to be used in figure titles
        and axes.
    xlim
        Limits for histogram plotting.

    Returns
    -------
        Matplotlib figures containing histograms for each dataset within
        ``results`` and named variables in ``indexes``.
    """

    retval = []

    for k, v in results.items():
        samples = [s for s in v if len(s) != 2]
        for idx, name in indexes.items():
            val = numpy.array([s[idx] for s in samples])
            retval.append(
                make_histogram(
                    name, val, xlim=xlim, title=f"{name} Frequency Histogram (@ {k})"
                )
            )

    return retval



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""Tests for TB-POC dataset."""

import importlib

import pytest
from click.testing import CliRunner


def id_function(val):
    if isinstance(val, dict):
        return str(val)
    return repr(val)


@pytest.mark.parametrize(
    "split,lengths",
    [
        ("fold-0", dict(train=292, validation=74, test=41)),
        ("fold-1", dict(train=292, validation=74, test=41)),
        ("fold-2", dict(train=292, validation=74, test=41)),
        ("fold-3", dict(train=292, validation=74, test=41)),
        ("fold-4", dict(train=292, validation=74, test=41)),
        ("fold-5", dict(train=292, validation=74, test=41)),
        ("fold-6", dict(train=292, validation=74, test=41)),
        ("fold-7", dict(train=293, validation=74, test=40)),
        ("fold-8", dict(train=293, validation=74, test=40)),
        ("fold-9", dict(train=293, validation=74, test=40)),
    ],
    ids=id_function,  # just changes how pytest prints it
)
def test_protocol_consistency(
    database_checkers,
    split: str,
    lengths: dict[str, int],
):
    from mednet.data.split import make_split

    database_checkers.check_split(
        make_split("mednet.config.classify.data.tbpoc", f"{split}.json"),
        lengths=lengths,
        prefixes=(
            "TBPOC_CXR/TBPOC-",
            "TBPOC_CXR/tbpoc-",
        ),
        possible_labels=(0, 1),
    )


@pytest.mark.skip_if_rc_var_not_set("datadir.tbpoc")
def test_database_check():
    from mednet.scripts.database import check

    runner = CliRunner()
    result = runner.invoke(check, ["tbpoc-f0"])
    assert (
        result.exit_code == 0
    ), f"Exit code {result.exit_code} != 0 -- Output:\n{result.output}"


@pytest.mark.skip_if_rc_var_not_set("datadir.tbpoc")
@pytest.mark.parametrize(
    "dataset",
    [
        "train",
        "validation",
        "test",
    ],
)
@pytest.mark.parametrize(
    "name",
    [
        "fold_0",
        "fold_1",
        "fold_2",
        "fold_3",
        "fold_4",
        "fold_5",
        "fold_6",
        "fold_7",
        "fold_8",
        "fold_9",
    ],
)
def test_loading(database_checkers, name: str, dataset: str):
    datamodule = importlib.import_module(
        f".{name}", "mednet.config.classify.data.tbpoc"
    ).datamodule

    datamodule.model_transforms = []  # should be done before setup()
    datamodule.setup("predict")  # sets up all datasets

    loader = datamodule.predict_dataloader()[dataset]

    limit = 3  # limit load checking
    for batch in loader:
        if limit == 0:
            break
        database_checkers.check_loaded_batch(
            batch,
            batch_size=1,
            color_planes=1,
            prefixes=("TBPOC_CXR/TBPOC-", "TBPOC_CXR/tbpoc-"),
            possible_labels=(0, 1),
            expected_num_labels=1,
            expected_meta_size=2,
        )
        limit -= 1


@pytest.mark.skip_if_rc_var_not_set("datadir.tbpoc")
def test_loaded_image_quality(database_checkers, datadir):
    reference_histogram_file = (
        datadir / "histograms/raw_data/histograms_tbpoc_fold_0.json"
    )

    datamodule = importlib.import_module(
        ".fold_0", "mednet.config.classify.data.tbpoc"
    ).datamodule

    datamodule.model_transforms = []
    datamodule.setup("predict")

    database_checkers.check_image_quality(datamodule, reference_histogram_file)



# SPDX-FileCopyrightText: Copyright © 2024 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later


from mednet.engine.detect.evaluator import run
from mednet.models.detect.typing import Prediction


def test_iou_from_no_predictions():
    predictions: list[Prediction] = [
        (
            "path/to/sample1.png",
            # bounding-box, target
            (([10, 10, 20, 20], 1),),  # target #1
            (),  # no predictions
        )
    ]

    result = run(predictions, binning=10)

    assert result["num-samples"] == 1
    assert result["num-targets"] == 1
    assert result["num-detections"] == 0
    assert result["mean-iou"] == 0.0
    assert len(result["per-class"]) == 1
    assert 1 in result["per-class"]
    assert result["per-class"][1]["mean-iou"] == 0.0
    assert result["per-class"][1]["num-targets"] == 1
    assert result["per-class"][1]["num-detections"] == 0


def test_iou_from_no_matching_predictions():
    predictions: list[Prediction] = [
        (
            "path/to/sample1.png",
            # ground-truth: bounding-box, target
            (([10, 10, 20, 20], 1),),  # target #1
            # detections: bounding-box, target, classifier score
            (([10, 10, 20, 20], 0, 1.0),),  # does not match target
        )
    ]

    result = run(predictions, binning=10)

    assert result["num-samples"] == 1
    assert result["num-targets"] == 1
    assert result["num-detections"] == 1
    assert result["mean-iou"] == 0.0
    assert len(result["per-class"]) == 1
    assert 1 in result["per-class"]
    assert result["per-class"][1]["mean-iou"] == 0.0
    assert result["per-class"][1]["num-targets"] == 1
    assert result["per-class"][1]["num-detections"] == 0


def test_iou_from_single_prediction():
    predictions: list[Prediction] = [
        (
            "path/to/sample1.png",
            # ground-truth: bounding-box, target
            (([10, 10, 20, 20], 1),),  # target #1
            # detections: bounding-box, target, classifier score
            (([10, 10, 20, 20], 1, 1.0),),  # perfectly matches target
        )
    ]

    result = run(predictions, binning=10)

    assert result["num-samples"] == 1
    assert result["num-targets"] == 1
    assert result["num-detections"] == 1
    assert result["mean-iou"] == 1.0
    assert len(result["per-class"]) == 1
    assert 1 in result["per-class"]
    assert result["per-class"][1]["mean-iou"] == 1.0
    assert result["per-class"][1]["num-targets"] == 1
    assert result["per-class"][1]["num-detections"] == 1


def test_iou_from_multiple_predictions():
    predictions: list[Prediction] = [
        (
            "path/to/sample1.png",
            # ground-truth: bounding-box, target
            (([10, 10, 20, 20], 1),),  # target #1
            # detections: bounding-box, target, classifier score
            (
                ([10, 10, 20, 20], 0, 0.1),  # does not match target
                ([15, 10, 20, 20], 1, 0.9),  # somewhat matches target
            ),
        )
    ]

    result = run(predictions, binning=10)

    assert result["num-samples"] == 1
    assert result["num-targets"] == 1
    assert result["num-detections"] == 2
    assert result["mean-iou"] == 0.5
    assert len(result["per-class"]) == 1
    assert 1 in result["per-class"]
    assert result["per-class"][1]["mean-iou"] == 0.5
    assert result["per-class"][1]["num-targets"] == 1
    assert result["per-class"][1]["num-detections"] == 1


def test_iou_from_multiple_predictions_that_can_match():
    predictions: list[Prediction] = [
        (
            "path/to/sample1.png",
            # ground-truth: bounding-box, target
            (([10, 10, 20, 20], 1),),  # target #1
            # detections: bounding-box, target, classifier score
            (
                ([15, 10, 20, 20], 1, 0.9),  # somewhat matches target
                ([12.5, 10, 20, 20], 1, 0.9),  # better matches target
            ),
        )
    ]

    result = run(predictions, binning=10)

    assert result["num-samples"] == 1
    assert result["num-targets"] == 1
    assert result["num-detections"] == 2
    assert result["mean-iou"] == 0.75
    assert len(result["per-class"]) == 1
    assert 1 in result["per-class"]
    assert result["per-class"][1]["mean-iou"] == 0.75
    assert result["per-class"][1]["num-targets"] == 1
    assert result["per-class"][1]["num-detections"] == 2


def test_iou_from_prediction_that_misses():
    predictions: list[Prediction] = [
        (
            "path/to/sample1.png",
            # ground-truth: bounding-box, target
            (([10, 10, 20, 20], 1),),  # target #1
            # detections: bounding-box, target, classifier score
            (
                ([30, 30, 40, 40], 1, 0.2),  # completely misses
            ),
        )
    ]

    result = run(predictions, binning=10)

    assert result["num-samples"] == 1
    assert result["num-targets"] == 1
    assert result["num-detections"] == 1
    assert result["mean-iou"] == 0.0
    assert len(result["per-class"]) == 1
    assert 1 in result["per-class"]
    assert result["per-class"][1]["mean-iou"] == 0.0
    assert result["per-class"][1]["num-targets"] == 1
    assert result["per-class"][1]["num-detections"] == 1


def test_iou_from_multiple_predictions_and_targets():
    predictions: list[Prediction] = [
        (
            "path/to/sample1.png",
            # ground-truth: bounding-box, target
            (
                ([10, 10, 20, 20], 1),  # target #0
                ([30, 30, 40, 40], 1),  # target #1
            ),
            # detections: bounding-box, target, classifier score
            (
                ([35, 30, 40, 40], 1, 0.9),  # matches target #1
                ([12.5, 10, 20, 20], 1, 0.9),  # matches target #0
            ),
        )
    ]

    result = run(predictions, binning=10)

    assert result["num-samples"] == 1
    assert result["num-targets"] == 2
    assert result["num-detections"] == 2
    assert result["mean-iou"] == (0.75 + 0.5) / 2
    assert len(result["per-class"]) == 1
    assert 1 in result["per-class"]
    assert result["per-class"][1]["mean-iou"] == (0.75 + 0.5) / 2
    assert result["per-class"][1]["num-targets"] == 2
    assert result["per-class"][1]["num-detections"] == 2


def test_iou_from_multiple_samples():
    predictions: list[Prediction] = [
        (
            "path/to/sample1.png",
            # ground-truth: bounding-box, target
            (([10, 10, 20, 20], 1),),
            # detections: bounding-box, target, classifier score
            (([12.5, 10, 20, 20], 1, 0.9),),
        ),
        (
            "path/to/sample2.png",
            # ground-truth: bounding-box, target
            (([30, 30, 40, 40], 1),),
            # detections: bounding-box, target, classifier score
            (
                ([35, 30, 40, 40], 1, 0.9),  # matches target
            ),
        ),
    ]

    result = run(predictions, binning=10)

    assert result["num-samples"] == 2
    assert result["num-targets"] == 2
    assert result["num-detections"] == 2
    assert result["mean-iou"] == (0.75 + 0.5) / 2
    assert len(result["per-class"]) == 1
    assert 1 in result["per-class"]
    assert result["per-class"][1]["mean-iou"] == (0.75 + 0.5) / 2
    assert result["per-class"][1]["num-targets"] == 2
    assert result["per-class"][1]["num-detections"] == 2


def test_iou_from_multiple_samples_and_classes():
    predictions: list[Prediction] = [
        (
            "path/to/sample1.png",
            # ground-truth: bounding-box, target
            (([10, 10, 20, 20], 1),),
            # detections: bounding-box, target, classifier score
            (([12.5, 10, 20, 20], 1, 0.9),),
        ),
        (
            "path/to/sample2.png",
            # ground-truth: bounding-box, target
            (([30, 30, 40, 40], 2),),
            # detections: bounding-box, target, classifier score
            (
                ([35, 30, 40, 40], 2, 0.9),  # matches target
            ),
        ),
    ]

    result = run(predictions, binning=10)

    assert result["num-samples"] == 2
    assert result["num-targets"] == 2
    assert result["num-detections"] == 2
    assert result["mean-iou"] == (0.75 + 0.5) / 2
    assert len(result["per-class"]) == 2
    assert 1 in result["per-class"]
    assert result["per-class"][1]["mean-iou"] == 0.75
    assert result["per-class"][1]["num-targets"] == 1
    assert result["per-class"][1]["num-detections"] == 1
    assert result["per-class"][2]["mean-iou"] == 0.5
    assert result["per-class"][2]["num-targets"] == 1
    assert result["per-class"][2]["num-detections"] == 1



# Copyright © 2022 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import os

import click
from clapper.click import ConfigCommand as _BaseConfigCommand


class ConfigCommand(_BaseConfigCommand):
    """A click command-class that has the properties of :py:class:`clapper.click.ConfigCommand` and adds verbatim epilog formatting."""

    def format_epilog(self, ctx: click.Context, formatter: click.HelpFormatter) -> None:
        """Format the command epilog during --help.

        Parameters
        ----------
        ctx
            The current parsing context.
        formatter
            The formatter to use for printing text.
        """
        del ctx  # silence linter

        if self.epilog:
            formatter.write_paragraph()
            for line in self.epilog.split("\n"):
                formatter.write(line + "\n")


class PathOrURL(click.Path):
    """A click parameter type that represents either an URL or a filesystem path.

    This object may be initialized just like :py:class:`click.Path`. If the
    input resembles an URL (starts with "http"), then its returned as a string.
    Otherwise, it is returned depending on the way it was initialized.

    Parameters
    ----------
    *args
        Positional arguments delivered to super class.
    **kwargs
        Keyword arguments delivered to super class.
    """

    name = "path_or_url"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def convert(
        self,
        value: str | os.PathLike[str],
        param: click.Parameter | None,
        ctx: click.Context | None,
    ) -> str | bytes | os.PathLike[str]:
        """Convert input strings into parameters.

        Parameters
        ----------
        value
            The value passed through the command-line.
        param
            The type of parameter.
        ctx
            The current parser context.

        Returns
        -------
            The parsed parameter.
        """

        if isinstance(value, str | bytes) and value.startswith("http"):
            return value

        return super().convert(value, param, ctx)



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import pathlib

import h5py
import numpy
import numpy.typing
import PIL.Image
import PIL.ImageOps
import torch
import torchvision.transforms.v2.functional

from .evaluator import tfpn_masks


def view(
    basedir: pathlib.Path,
    stem: str,
    threshold: float,
    show_errors: bool,
    tp_color: tuple[int, int, int],
    fp_color: tuple[int, int, int],
    fn_color: tuple[int, int, int],
    alpha: float,
) -> PIL.Image.Image:
    """Create an segmentation map visualisation.

    Parameters
    ----------
    basedir
        Base directory where the prediction indicated by ``stem`` is stored.
    stem
        Name of the HDF5 file containing the predictions, as output by the
        ``predict`` CLI.
    threshold
        The threshold to apply to the probability map loaded from the HDF5
        file.
    show_errors
        If set to ``True``, then colours false-positives (in red), and false
        negatives (in green).
    tp_color
        Tuple that indicates which color to use for displaying true-positives.
    fp_color
        Tuple that indicates which color to use for displaying false-positives.
    fn_color
        Tuple that indicates which color to use for displaying false-negatives.
    alpha
        How transparent will the overlay be.

    Returns
    -------
        An image with an overlayed segmentation map that can be saved or
        displayed.
    """

    def _to_pil(arr: numpy.typing.NDArray[numpy.float32]) -> PIL.Image.Image:
        return torchvision.transforms.v2.functional.to_pil_image(torch.Tensor(arr))

    with h5py.File(basedir / stem, "r") as f:
        image: numpy.typing.NDArray[numpy.float32] = numpy.array(f["image"])
        target: numpy.typing.NDArray[numpy.bool_] = numpy.array(f["target"])
        mask: numpy.typing.NDArray[numpy.bool_] = numpy.array(f["mask"])

        pred: numpy.typing.NDArray[numpy.float32] = numpy.empty(
            (0,), dtype=numpy.float32
        )
        if "prediction" in f:
            pred = numpy.array(f["prediction"])

    image *= mask
    target = numpy.logical_and(target, mask)

    if pred.shape == (0,):
        # no prediction available, can only show target

        image *= mask
        target = numpy.logical_and(target, mask)

        overlay = target >= 0.5
        tp_pil = _to_pil(overlay.astype(numpy.float32))
        tp_pil_colored = PIL.ImageOps.colorize(tp_pil, (0, 0, 0), tp_color)

        retval = _to_pil(image)
        return PIL.Image.blend(retval, tp_pil_colored, alpha)

    # there is a prediction - show more information
    pred *= mask
    if show_errors:
        tp, fp, _, fn = tfpn_masks(pred, target, threshold)

        # change to PIL representation
        tp_pil = _to_pil(tp.astype(numpy.float32))
        tp_pil_colored = PIL.ImageOps.colorize(tp_pil, (0, 0, 0), tp_color)

        fp_pil = _to_pil(fp.astype(numpy.float32))
        fp_pil_colored = PIL.ImageOps.colorize(fp_pil, (0, 0, 0), fp_color)

        fn_pil = _to_pil(fn.astype(numpy.float32))
        fn_pil_colored = PIL.ImageOps.colorize(fn_pil, (0, 0, 0), fn_color)

        tp_pil_colored.paste(fp_pil_colored, mask=fp_pil)
        tp_pil_colored.paste(fn_pil_colored, mask=fn_pil)

    else:
        overlay = pred >= threshold
        tp_pil = _to_pil(overlay.astype(numpy.float32))
        tp_pil_colored = PIL.ImageOps.colorize(tp_pil, (0, 0, 0), tp_color)

    retval = _to_pil(image)
    return PIL.Image.blend(retval, tp_pil_colored, alpha)



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import pathlib

import clapper.click
import click

from ..utils.string import rewrap
from .click import ConfigCommand, PathOrURL
from .logging import setup_cli_logger

logger = setup_cli_logger()


@click.command(
    entry_point_group="mednet.config",
    cls=ConfigCommand,
    epilog="""Examples:

1. Run prediction on an existing DataModule configuration:

   .. code:: sh

      mednet predict -vv lwnet drive --weight=path/to/model.ckpt --output-folder=path/to/predictions

2. Enable multi-processing data loading with 6 processes:

   .. code:: sh

      mednet predict -vv lwnet drive --parallel=6 --weight=path/to/model.ckpt --output-folder=path/to/predictions

""",
)
@click.option(
    "--output-folder",
    "-o",
    help="Directory in which to save predictions (created if does not exist)",
    required=True,
    type=click.Path(
        file_okay=False,
        dir_okay=True,
        writable=True,
        path_type=pathlib.Path,
    ),
    default="results",
    show_default=True,
    cls=clapper.click.ResourceOption,
)
@click.option(
    "--model",
    "-m",
    help="""A lightning module instance implementing the network architecture
    (not the weights, necessarily) to be used for prediction.""",
    required=True,
    cls=clapper.click.ResourceOption,
)
@click.option(
    "--datamodule",
    "-d",
    help=rewrap(
        """A lightning DataModule that will be asked for prediction data loaders.
            Typically, this includes all configured splits in a DataModule, however this
            is not a requirement.  A DataModule that returns a single dataloader for
            prediction (wrapped in a dictionary) is acceptable."""
    ),
    required=True,
    cls=clapper.click.ResourceOption,
)
@click.option(
    "--batch-size",
    "-b",
    help=rewrap(
        """Number of samples in every batch (this parameter affects memory requirements
        for the network)."""
    ),
    required=True,
    show_default=True,
    default=1,
    type=click.IntRange(min=1),
    cls=clapper.click.ResourceOption,
)
@click.option(
    "--device",
    "-d",
    help=rewrap(
        """A string indicating the device to use (e.g. "cpu", "cuda", "mps",
        "cuda:0"). Can be passed as a string, or as an environment variable (e.g.
        `MEDNET_DEVICE=cpu` or `MEDNET_DEVICE=mps`)."""
    ),
    envvar="MEDNET_DEVICE",
    show_default=True,
    required=True,
    default="cpu",
    cls=clapper.click.ResourceOption,
)
@click.option(
    "--weight",
    "-w",
    help=rewrap(
        """Path or URL to pretrained model file (`.ckpt` extension), corresponding to
        the architecture set with `--model`.  Optionally, you may also pass a directory
        containing the result of a training session, in which case either the best
        (lowest validation) or latest model will be loaded."""
    ),
    required=True,
    cls=clapper.click.ResourceOption,
    type=PathOrURL(
        exists=True,
        file_okay=True,
        dir_okay=True,
        readable=True,
        path_type=pathlib.Path,
    ),
)
@click.option(
    "--parallel",
    "-P",
    help=rewrap(
        """Use multiprocessing for data loading: if set to -1 (default), disables
        multiprocessing data loading.  Set to 0 to enable as many data loading instances
        as processing cores available in the system.  Set to >= 1 to enable that many
        multiprocessing instances for data loading."""
    ),
    type=click.IntRange(min=-1),
    show_default=True,
    required=True,
    default=-1,
    cls=clapper.click.ResourceOption,
)
@clapper.click.verbosity_option(logger=logger, expose_value=False)
def predict(
    output_folder,
    model,
    datamodule,
    batch_size,
    device,
    weight,
    parallel,
    **_,
) -> None:  # numpydoc ignore=PR01
    """Run inference on input samples, using a pre-trained model."""

    from ..engine.device import DeviceManager
    from ..engine.trainer import validate_model_datamodule
    from ..utils.checkpointer import get_checkpoint_to_run_inference
    from .utils import JSONable, save_json_metadata, save_json_with_backup

    validate_model_datamodule(model, datamodule)

    # sets-up the data module
    datamodule.batch_size = batch_size
    datamodule.parallel = parallel
    datamodule.model_transforms = list(model.model_transforms)

    datamodule.prepare_data()
    datamodule.setup(stage="predict")

    if isinstance(weight, pathlib.Path) and weight.is_dir():
        weight = get_checkpoint_to_run_inference(weight)

    logger.info(f"Loading checkpoint from `{weight}`...")
    model = type(model).load_from_checkpoint(weight, strict=False)

    device_manager = DeviceManager(device)

    save_json_metadata(
        output_file=output_folder / "predictions.meta.json",
        output_folder=output_folder,
        model=model,
        datamodule=datamodule,
        batch_size=batch_size,
        device=device,
        weight=weight,
        parallel=parallel,
    )

    predictions: JSONable = None
    match datamodule.task:
        case "classification":
            from ..engine.classify.predictor import run as run_classify

            logger.info(f"Running prediction for `{datamodule.task}` task...")
            predictions = run_classify(model, datamodule, device_manager)

        case "segmentation":
            from ..engine.segment.predictor import run as run_segment

            logger.info(f"Running prediction for `{datamodule.task}` task...")
            predictions = run_segment(model, datamodule, device_manager, output_folder)

        case "detection":
            from ..engine.detect.predictor import run as run_detect

            logger.info(f"Running prediction for `{datamodule.task}` task...")
            predictions = run_detect(model, datamodule, device_manager)

        case _:
            raise click.BadParameter(
                f"Do not know how to handle `{datamodule.task}` task from "
                f"`{type(datamodule).__module__}.{type(datamodule).__name__}`"
            )

    predictions_file = output_folder / "predictions.json"
    save_json_with_backup(predictions_file, predictions)
    logger.info(f"Predictions saved to `{str(predictions_file)}`")



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""Test code for database splits."""

from mednet.data.split import JSONDatabaseSplit


def test_json_loading(datadir):
    # tests if we can build a simple JSON loader for the Iris Flower dataset

    database_split = JSONDatabaseSplit(datadir / "iris.json")

    assert len(database_split["train"]) == 75
    for k in database_split["train"]:
        for f in range(4):
            assert isinstance(k[f], int | float)
        assert isinstance(k[4], str)

    assert len(database_split["test"]) == 75
    for k in database_split["test"]:
        for f in range(4):
            assert isinstance(k[f], int | float)
        assert isinstance(k[4], str)



# SPDX-FileCopyrightText: Copyright © 2024 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""DRIONS-DB (training set) for Optic Disc Segmentation.

The dataset originates from data collected from 55 patients with glaucoma
(23.1%) and eye hypertension (76.9%), and random selected from an eye fundus
image base belonging to the Ophthalmology Service at Miguel Servet Hospital,
Saragossa (Spain).  It contains 110 eye fundus images with a resolution of 600
x 400. Two sets of ground-truth optic disc annotations are available. The first
set is commonly used for training and testing. The second set acts as a "human"
baseline.

* Reference: :cite:p:`carmona_identification_2008`
* Original resolution (height x width): 400 x 600
* Configuration resolution: 416 x 608 (after padding)
* Split reference: :cite:p:`maninis_deep_2016`
* Protocols ``expert1`` (baseline) and ``expert2`` (human comparison):

  * Training samples: 60
  * Test samples: 50

This module contains the base declaration of common data modules and raw-data
loaders for this database. All configured splits inherit from this definition.
"""

import csv
import importlib.resources
import importlib.resources.abc
import os
import pathlib
import typing

import PIL.Image
import PIL.ImageDraw
import torch
from torchvision import tv_tensors
from torchvision.transforms.v2.functional import to_dtype, to_image

from ...models.transforms import crop_image_to_mask
from ...utils.rc import load_rc
from ..datamodule import CachingDataModule
from ..split import JSONDatabaseSplit
from ..typing import RawDataLoader as BaseDataLoader
from ..typing import Sample

DATABASE_SLUG = __name__.rsplit(".", 1)[-1]
"""Pythonic name to refer to this database."""

CONFIGURATION_KEY_DATADIR = "datadir." + DATABASE_SLUG
"""Key to search for in the configuration file for the root directory of this
database."""


class RawDataLoader(BaseDataLoader):
    """A specialized raw-data-loader for the drionsdb dataset."""

    datadir: pathlib.Path
    """This variable contains the base directory where the database raw data is
    stored."""

    def __init__(self):
        self.datadir = pathlib.Path(
            load_rc().get(CONFIGURATION_KEY_DATADIR, os.path.realpath(os.curdir))
        )

    def _txt_to_pil_1(
        self, fname: pathlib.Path, size: tuple[int, int]
    ) -> PIL.Image.Image:
        """Convert DRIONS-DB annotations to image format.

        Parameters
        ----------
        fname
            Path to a file containing annotations.
        size
            The size of the mask (width, height).

        Returns
        -------
            The binary mask.
        """
        with fname.open("r") as f:
            rows = csv.reader(f, delimiter=",", quoting=csv.QUOTE_NONNUMERIC)
            data = list(map(tuple, rows))

        retval = PIL.Image.new("1", size)
        draw = PIL.ImageDraw.ImageDraw(retval)
        draw.polygon(data, fill="white")
        del draw
        return retval

    def sample(self, sample: typing.Any) -> Sample:
        """Load a single image sample from the disk.

        Parameters
        ----------
        sample
            A tuple containing path suffixes to the sample image, target, and mask
            to be loaded, within the dataset root folder.

        Returns
        -------
            The sample representation.
        """

        image = PIL.Image.open(self.datadir / sample[0]).convert(mode="RGB")
        image = to_dtype(to_image(image), torch.float32, scale=True)

        target = self.target(sample)

        mask_path = (
            importlib.resources.files(__package__) / "masks" / DATABASE_SLUG / sample[2]
        )
        with importlib.resources.as_file(mask_path) as path:
            mask = PIL.Image.open(path).convert(mode="1", dither=None)
            mask = to_dtype(to_image(mask), torch.float32, scale=True)

        image = tv_tensors.Image(crop_image_to_mask(image, mask))
        target = tv_tensors.Mask(crop_image_to_mask(target, mask))
        mask = tv_tensors.Mask(mask)

        return dict(image=image, target=target, mask=mask, name=sample[0])

    def target(self, sample: typing.Any) -> torch.Tensor:
        """Load only sample target from its raw representation.

        Parameters
        ----------
        sample
            A tuple containing the path suffix, within the dataset root folder,
            where to find the image to be loaded, and an integer, representing
            the sample target.

        Returns
        -------
            The label corresponding to the specified sample, encapsulated as a
            torch float tensor.
        """

        image = PIL.Image.open(self.datadir / sample[0])
        target = self._txt_to_pil_1(self.datadir / sample[1], image.size)
        return to_dtype(to_image(target), torch.float32, scale=True)


class DataModule(CachingDataModule):
    """DRIONS-DB (training set) for Optic Disc Segmentation.

    Parameters
    ----------
    split_path
        Path or traversable (resource) with the JSON split description to load.
    """

    def __init__(self, split_path: pathlib.Path | importlib.resources.abc.Traversable):
        super().__init__(
            database_split=JSONDatabaseSplit(split_path),
            raw_data_loader=RawDataLoader(),
            database_name=DATABASE_SLUG,
            split_name=split_path.name.rsplit(".", 2)[0],
            task="segmentation",
            num_classes=1,
        )



# SPDX-FileCopyrightText: Copyright © 2024 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""Little W-Net for image segmentation.

The Little W-Net architecture contains roughly around 70k parameters and
closely matches (or outperforms) other more complex techniques.

Reference: :cite:p:`galdran_state---art_2022`
"""

import torch.optim
import torch.optim.lr_scheduler
import torchvision.transforms
import torchvision.transforms.v2

import mednet.models.losses
import mednet.models.segment.lwnet
import mednet.models.transforms

max_lr = 0.01  # start
min_lr = 1e-08  # valley
# Original strategy: https://github.com/agaldran/lwnet/blob/master/train_cyclical.py#L298
# About 20 * len(train-data-loader)
cycle = 100  # 1/2 epochs for a complete scheduling cycle

model = mednet.models.segment.lwnet.LittleWNet(
    loss_type=mednet.models.losses.MultiLayerBCELogitsLossWeightedPerBatch,
    loss_arguments=dict(),
    optimizer_type=torch.optim.Adam,
    optimizer_arguments=dict(lr=max_lr),
    scheduler_type=torch.optim.lr_scheduler.CosineAnnealingLR,
    scheduler_arguments=dict(T_max=cycle, eta_min=min_lr),
    model_transforms=[
        mednet.models.transforms.SquareCenterPad(),
        torchvision.transforms.v2.Resize(512, antialias=True),
        torchvision.transforms.v2.RGB(),
    ],
)



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import pathlib

from tensorboard.backend.event_processing.event_accumulator import (
    EventAccumulator,
)


def scalars_to_dict(
    logdir: pathlib.Path,
) -> dict[str, tuple[list[int], list[float]]]:
    """Return scalars stored in tensorboard event files.

    This method will gather all tensorboard event files produced by a training
    run, and will return a dictionary with all collected scalars, ready for
    plotting.

    Parameters
    ----------
    logdir
        Directory containing the event files.

    Returns
    -------
    dict[str, tuple[list[int], list[float]]]
        A dictionary where keys represent all scalar names, and values
        correspond to a tuple that contains an array with epoch numbers (when
        values were taken), when the monitored values themselves.  The lists
        are pre-sorted by epoch number.
    """

    retval: dict[str, tuple[list[int], list[float]]] = {}

    for logfile in sorted(logdir.glob("events.out.tfevents.*")):
        event_accumulator = EventAccumulator(str(logfile))
        event_accumulator.Reload()

        for tag in event_accumulator.Tags()["scalars"]:
            steps, values = retval.setdefault(tag, ([], []))
            for data_point in event_accumulator.Scalars(tag):
                steps.append(data_point.step)
                values.append(data_point.value)

    # reorder according to step number
    for key, (steps, values) in retval.items():
        _steps, _values = zip(*sorted(zip(steps, values)))
        retval[key] = (list(_steps), list(_values))  # type: ignore

    return retval



# Copyright © 2022 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import importlib

import click
from clapper.click import AliasedGroup


@click.group(
    cls=AliasedGroup,
    context_settings=dict(help_option_names=["-?", "-h", "--help"]),
)
def classify():
    """Image classification specialized commands."""
    pass


def _add_command(cli, module, obj):
    cli.add_command(
        getattr(importlib.import_module("." + module, package=__name__), obj)
    )


_add_command(classify, ".evaluate", "evaluate")


@click.group(
    cls=AliasedGroup,
    context_settings=dict(help_option_names=["-?", "-h", "--help"]),
)
def saliency():
    """Generate, evaluate and view saliency maps."""
    pass


classify.add_command(saliency)

_add_command(saliency, ".saliency.generate", "generate")
_add_command(saliency, ".saliency.completeness", "completeness")
_add_command(saliency, ".saliency.interpretability", "interpretability")
_add_command(saliency, ".saliency.view", "view")



# Copyright © 2022 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""Aggregated DataModule composed of :py:mod:`montgomery's <.data.classify.montgomery>`, :py:mod:`shenzhen's <.data.classify.shenzhen>`, :py:mod:`indian's <.data.classify.indian>`, and :py:mod:`tbx11k's <.data.classify.tbx11k>` splits.

This module contains the base declaration of common data modules and raw-data
loaders for this database. All configured splits inherit from this definition.
"""

import importlib.resources.abc
import pathlib

from ..datamodule import ConcatDataModule
from ..split import JSONDatabaseSplit
from .indian import CONFIGURATION_KEY_DATADIR as INDIAN_KEY_DATADIR
from .indian import RawDataLoader as IndianLoader
from .montgomery import RawDataLoader as MontgomeryLoader
from .shenzhen import RawDataLoader as ShenzhenLoader
from .tbx11k import RawDataLoader as TBX11kLoader

DATABASE_SLUG = __name__.rsplit(".", 1)[-1]
"""Pythonic name of this database."""


class DataModule(ConcatDataModule):
    """Aggregated DataModule composed of :py:mod:`montgomery's <.data.classify.montgomery>`, :py:mod:`shenzhen's <.data.classify.shenzhen>`, :py:mod:`indian's <.data.classify.indian>`, and :py:mod:`tbx11k's <.data.classify.tbx11k>` splits.

    Parameters
    ----------
    split_name
        The name of the split to assign to this data module.
    split_path
        Path or traversable (resource) with the JSON split description to load
        for montgomery, shenzhen, indian and tbx11k databases (in this order).
    """

    def __init__(
        self,
        split_name: str,
        split_path: tuple[
            pathlib.Path | importlib.resources.abc.Traversable,
            pathlib.Path | importlib.resources.abc.Traversable,
            pathlib.Path | importlib.resources.abc.Traversable,
            pathlib.Path | importlib.resources.abc.Traversable,
        ],
    ):
        montgomery_loader = MontgomeryLoader()
        montgomery_split = JSONDatabaseSplit(split_path[0])
        shenzhen_loader = ShenzhenLoader()
        shenzhen_split = JSONDatabaseSplit(split_path[1])
        indian_loader = IndianLoader(INDIAN_KEY_DATADIR)
        indian_split = JSONDatabaseSplit(split_path[2])
        tbx11k_loader = TBX11kLoader(ignore_bboxes=True)
        tbx11k_split = JSONDatabaseSplit(split_path[3])

        super().__init__(
            splits={
                "train": [
                    (montgomery_split["train"], montgomery_loader),
                    (shenzhen_split["train"], shenzhen_loader),
                    (indian_split["train"], indian_loader),
                    (tbx11k_split["train"], tbx11k_loader),
                ],
                "validation": [
                    (montgomery_split["validation"], montgomery_loader),
                    (shenzhen_split["validation"], shenzhen_loader),
                    (indian_split["validation"], indian_loader),
                    (tbx11k_split["validation"], tbx11k_loader),
                ],
                "test": [
                    (montgomery_split["test"], montgomery_loader),
                    (shenzhen_split["test"], shenzhen_loader),
                    (indian_split["test"], indian_loader),
                    (tbx11k_split["test"], tbx11k_loader),
                ],
            },
            database_name=DATABASE_SLUG,
            split_name=split_name,
            task="classification",
            num_classes=1,
        )
